'''
Created Date: Friday May 31st 2024 +1000
Author: Peter Baker
-----
Last Modified: Friday May 31st 2024 9:50:26 am +1000
Modified By: Peter Baker
-----
Description: The main Provena L3 client class which collects sub L2 clients and L3 clients.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Cleaned up and added missing modules.
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import *
from provenaclient.modules import *
from provenaclient.modules.module_helpers import *

# L3 interface.


class ProvenaClient(ModuleService):
    # L2 + L3 combinations

    # Data store
    _datastore_client: DatastoreClient
    datastore: Datastore
    
    # Search 
    _search_client: SearchClient
    search: Search
    
    # Auth 
    _auth_client: AuthClient
    auth_api: Auth
    
    # Registry 
    _registry_client: RegistryClient
    registry: Registry
    
    # Prov 
    _prov_client: ProvClient
    prov_api: Prov
    
    # Jobs 
    _job_client: JobAPIClient
    job_api: JobService
    
    # ID service 
    _id_client: IdServiceClient
    id_api: IDService

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """
        
        Build an instance of the Provena Client.

        Args:
            auth (AuthManager): The Auth implementation to use. See auth.implementations
            config (Config): The provena config which indicates deployment and other settings
        """
        # Module service
        self._auth = auth
        self._config = config

        # (L2 clients)
        self._datastore_client = DatastoreClient(auth, config)
        self._search_client = SearchClient(auth, config)
        self._auth_client = AuthClient(auth, config)
        self._registry_client = RegistryClient(auth, config)
        self._prov_client = ProvClient(auth, config)
        self._job_client = JobAPIClient(auth,config)
        self._id_client = IdServiceClient(auth,config)

        self.datastore = Datastore(
            auth=auth,
            config=config,
            datastore_client=self._datastore_client,
            search_client=self._search_client
        )

        self.search = Search(
            auth=auth,
            config=config,
            search_client=self._search_client
        )

        self.auth_api = Auth(
            auth=auth,
            config=config,
            auth_client=self._auth_client
        )

        self.registry = Registry(
            auth=auth, 
            config=config,
            registry_client=self._registry_client
        )

        self.prov_api = Prov(
            auth=auth, 
            config=config, 
            prov_client=self._prov_client
        )
        
        self.job_api = JobService(
            auth=auth, 
            config=config, 
            job_api_client=self._job_client
        )
        
        self.id_api = IDService(
            auth=auth, 
            config=config, 
            id_service_client=self._id_client
        )


'''
Created Date: Friday May 31st 2024 +1000
Author: Peter Baker
-----
Last Modified: Friday May 31st 2024 9:50:26 am +1000
Modified By: Peter Baker
-----
Description: The ID Service L3 module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import IdServiceClient
from provenaclient.utils.exceptions import *
from provenaclient.modules.module_helpers import *
from ProvenaInterfaces.HandleAPI import *

# L3 interface.


class IDService(ModuleService):
    _id_service_client: IdServiceClient

    def __init__(self, auth: AuthManager, config: Config, id_service_client: IdServiceClient) -> None:
        """Initialises a new id-service object, which sits between the user and the id-service api operations.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance. 
        datastore_client : DatastoreClient
            This client interacts with the Datastore API's.
        """
        self._auth = auth
        self._config = config

        # Clients related to the handle-api scoped as private.
        self.id_service_client = id_service_client

    async def mint(self, body: MintRequest) -> Handle:
        """

        Mints a new handle with given properties.

        Args:
            body (MintRequest): The payload

        Returns:
            Handle: The resulting handle object
        """
        return await self._id_service_client.mint(body=body)

    async def add_value(self, body: AddValueRequest) -> Handle:
        """
        Adds value to a handle.

        Args:
            body (AddValueRequest): The value to add

        Returns:
            Handle: The handle object
        """
        return await self._id_service_client.add_value(body=body)

    async def add_value_by_index(self, body: AddValueIndexRequest) -> Handle:
        """
        Adds a value at specified index

        Args:
            body (AddValueIndexRequest): The value to add

        Returns:
            Handle: The handle object
        """
        return await self._id_service_client.add_value_by_index(body=body)

    async def list(self) -> ListResponse:
        """
        Lists all handles under domain.

        Returns:
            ListResponse: The response list
        """
        return await self._id_service_client.list()

    async def modify_by_index(self, body: ModifyRequest) -> Handle:
        """
        Modifies existing handle value at specified index

        Args:
            body (ModifyRequest): The request to modify

        Returns:
            Handle: The handle object
        """
        return await self._id_service_client.modify_by_index(body=body)

    async def remove_by_index(self, body: RemoveRequest) -> Handle:
        """

        Removes value at specified index

        Args:
            body (RemoveRequest): The removal request

        Returns:
            Handle: The handle object
        """
        return await self._id_service_client.remove_by_index(body=body)



'''
Created Date: Tuesday June 18th 2024 +1000
Author: Peter Baker
-----
Last Modified: Tuesday June 18th 2024 12:56:41 pm +1000
Modified By: Peter Baker
-----
Description: Datastore file IO sub module, includes file upload and download helpers
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | First implementation including download_all_files and upload_all_files methods
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import DatastoreClient
from ProvenaInterfaces.DataStoreAPI import *
from provenaclient.modules.module_helpers import *
import cloudpathlib.s3 as s3  # type: ignore


def setup_s3_client(creds: CredentialResponse) -> s3.S3Client:
    """
    Uses the datastore creds response to generate an s3 cloud path lib client
    with auth.

    Args:
        creds (CredentialResponse): The data store credentials response

    Returns:
        s3.S3Client: The s3 client ready to use
    """
    # create dict of creds
    creds_dict = creds.credentials.dict()
    # don't want to pass expiry through as it confused s3 cloud lib
    del creds_dict['expiry']
    # create client
    client = s3.S3Client(
        **creds_dict
    )
    return client


def print_file_info(file: s3.S3Path) -> None:
    """
    Pretty prints a file specifying file/directory.
    File := s3.S3Path from Cloudpathlib

    Args:
        file (s3.S3Path): The file to print 
    """
    if file.is_dir():
        print(f"Directory: {file}")
    else:
        print(f"File: {file}")


class IOSubModule(ModuleService):
    _datastore_client: DatastoreClient

    def __init__(self, auth: AuthManager, config: Config, datastore_client: DatastoreClient) -> None:
        """
        File IO datastore sub module to assist with download and upload of data store files.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        auth_client: AuthClient
            The instantiated auth client
        """
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._datastore_client = datastore_client

    async def download_all_files(
        self,
        destination_directory: str,
        dataset_id: str
    ) -> None:
        """
        Downloads all files to the destination path for a given dataset id.

        - Fetches info
        - Fetches creds
        - Uses s3 cloud path lib to download all files to specified location

        Args:
            destination_directory (str): The destination path to save files to - use a directory
            dataset_id (str): The ID of the dataset to download files for - ensure you have read access
        """
        # Fetch the dataset information
        dataset_information = await self._datastore_client.fetch_dataset(
            id=dataset_id
        )
        # Get the S3 location
        assert dataset_information.item is not None, f"Expected non None item from dataset fetch, details: {dataset_information.status.details}."
        s3_location = dataset_information.item.s3
        # Get read credentials
        read_creds = await self._datastore_client.generate_read_access_credentials(
            read_access_credentials=CredentialsRequest(
                dataset_id=dataset_id,
                console_session_required=False
            )
        )
        # setup s3 client
        client = setup_s3_client(creds=read_creds)
        # create path
        path = s3.S3Path(cloud_path=s3_location.s3_uri, client=client)
        # download to specified path
        path.download_to(destination_directory)
        # Release file handles
        path.__del__

    async def list_all_files(
        self,
        dataset_id: str,
        print_list: bool = False,
    ) -> List[s3.S3Path]:
        """
        Lists all files stored in the given dataset by ID.

        - Fetches info
        - Fetches creds
        - Uses s3 cloud path lib to list all files to specified location

        Args:
            dataset_id (str): The ID of the dataset to download files for - ensure you have read access
        """
        # Fetch the dataset information
        dataset_information = await self._datastore_client.fetch_dataset(
            id=dataset_id
        )
        # Get the S3 location
        assert dataset_information.item is not None, f"Expected non None item from dataset fetch, details: {dataset_information.status.details}."
        s3_location = dataset_information.item.s3
        # Get read credentials
        read_creds = await self._datastore_client.generate_read_access_credentials(
            read_access_credentials=CredentialsRequest(
                dataset_id=dataset_id,
                console_session_required=False
            )
        )
        # setup s3 client
        client = setup_s3_client(creds=read_creds)
        # create path
        path = s3.S3Path(cloud_path=s3_location.s3_uri, client=client)
        paths = []
        for path in path.glob("**/*"):
            paths.append(path)

        if print_list:
            for p in paths:
                print_file_info(p)

        # Release file handles
        path.__del__

        return paths

    async def upload_all_files(
        self,
        source_directory: str,
        dataset_id: str
    ) -> None:
        """
        Uploads all files in the source path to the specified dataset id's storage location.

        - Fetches info
        - Fetches creds
        - Uses s3 cloud path lib to upload all files to specified location

        Args:
            source_directory (str): The source path to upload files from - use a directory
            dataset_id (str): The ID of the dataset to upload files for - ensure you have write access
        """
        # Fetch the dataset information
        dataset_information = await self._datastore_client.fetch_dataset(
            id=dataset_id
        )
        # Get the S3 location
        assert dataset_information.item is not None, f"Expected non None item from dataset fetch, details: {dataset_information.status.details}."
        s3_location = dataset_information.item.s3
        # Get read credentials
        write_creds = await self._datastore_client.generate_write_access_credentials(
            write_access_credentials=CredentialsRequest(
                dataset_id=dataset_id,
                console_session_required=False
            )
        )
        # setup s3 client
        client = setup_s3_client(creds=write_creds)
        # create path
        path = s3.S3Path(cloud_path=s3_location.s3_uri, client=client)
        # upload
        path.upload_from(source_directory)
        # Release file handles
        path.__del__



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Job API L2 Client.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from enum import Enum
from provenaclient.utils.helpers import *
from provenaclient.clients.client_helpers import *
from provenaclient.models import *
from ProvenaInterfaces.AsyncJobAPI import *


class JobAPIAdminEndpoints(str, Enum):
    """An ENUM containing the job api endpoints."""
    POST_JOBS_ADMIN_LAUNCH = "/jobs/admin/launch"
    GET_JOBS_ADMIN_FETCH = "/jobs/admin/fetch"
    POST_JOBS_ADMIN_LIST = "/jobs/admin/list"
    POST_JOBS_ADMIN_LIST_BATCH = "/jobs/admin/list_batch"

    # not implemented
    GET_ADMIN_CONFIG = "/admin/config"
    GET_ADMIN_SENTRY_DEBUG = "/admin/sentry-debug"


class JobAPIEndpoints(str, Enum):
    """An ENUM containing the job api endpoints."""
    GET_JOBS_USER_FETCH = "/jobs/user/fetch"
    POST_JOBS_USER_LIST = "/jobs/user/list"
    POST_JOBS_USER_LIST_BATCH = "/jobs/user/list_batch"

    # Done
    GET_HEALTH_CHECK = "/"

    # not implemented
    GET_CHECK_ACCESS_CHECK_GENERAL_ACCESS = "/check-access/check-general-access"
    GET_CHECK_ACCESS_CHECK_ADMIN_ACCESS = "/check-access/check-admin-access"
    GET_CHECK_ACCESS_CHECK_READ_ACCESS = "/check-access/check-read-access"
    GET_CHECK_ACCESS_CHECK_WRITE_ACCESS = "/check-access/check-write-access"

# L2 interface.


class JobAPIAdminSubClient(ClientService):
    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the JobAPIAdminClient sub client with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_endpoint(self, endpoint: JobAPIAdminEndpoints) -> str:
        return self._config.jobs_service_api_endpoint + endpoint.value

    async def launch_job(self, request: AdminLaunchJobRequest) -> AdminLaunchJobResponse:
        """
        Launches a new job.

        Args:
            request (AdminLaunchJobRequest): Specified job parameters

        Returns:
            AdminLaunchJobResponse: The response
        """
        return await parsed_post_request(
            client=self,
            json_body=py_to_dict(request),
            params={},
            url=self._build_endpoint(
                JobAPIAdminEndpoints.POST_JOBS_ADMIN_LAUNCH),
            error_message=f"Failed to launch job of subtype {request.job_sub_type}.",
            model=AdminLaunchJobResponse
        )

    async def get_job(self, session_id: str) -> AdminGetJobResponse:
        """
        Fetches a job (any job since admin)

        Args:
            session_id (str): The session ID of job to fetch

        Returns:
            AdminGetJobResponse: The response
        """
        return await parsed_get_request(
            client=self,
            params={'session_id': session_id},
            url=self._build_endpoint(
                JobAPIAdminEndpoints.GET_JOBS_ADMIN_FETCH),
            error_message=f"Failed to fetch job (admin) for session id {session_id}",
            model=AdminGetJobResponse
        )

    async def list_jobs(self, list_jobs_request: AdminListJobsRequest) -> AdminListJobsResponse:
        """
        Lists all jobs.

        Args:
            list_jobs_request (AdminListJobsRequest): The request including pagination information.

        Returns:
            AdminListJobsResponse: The list of jobs response.
        """
        return await parsed_post_request(
            client=self,
            json_body=py_to_dict(list_jobs_request),
            params={},
            url=self._build_endpoint(
                JobAPIAdminEndpoints.POST_JOBS_ADMIN_LIST),
            error_message=f"Failed to list jobs for user (admin).",
            model=AdminListJobsResponse
        )

    async def list_jobs_in_batch(self, list_request: AdminListByBatchRequest) -> AdminListByBatchResponse:
        """
        List jobs by batch ID, returning a list of jobs in the batch.

        Args:
            list_request (AdminListByBatchRequest): The request including batch ID

        Returns:
            AdminListByBatchResponse: The list of jobs in the batch
        """
        return await parsed_post_request(
            client=self,
            json_body=py_to_dict(list_request),
            params={},
            url=self._build_endpoint(
                JobAPIAdminEndpoints.POST_JOBS_ADMIN_LIST_BATCH),
            error_message=f"Failed to list jobs for specified batch (admin).",
            model=AdminListByBatchResponse
        )


class JobAPIClient(ClientService):
    admin: JobAPIAdminSubClient

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the JobAPIClient with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

        self.admin = JobAPIAdminSubClient(auth=auth, config=config)

    def _build_endpoint(self, endpoint: JobAPIEndpoints) -> str:
        return self._config.jobs_service_api_endpoint + endpoint.value

    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns:
            HealthCheckResponse: Response
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(JobAPIEndpoints.GET_HEALTH_CHECK),
            error_message="Health check failed!",
            params={},
            model=HealthCheckResponse
        )

    async def fetch_job(self, session_id: str) -> GetJobResponse:
        """
        Fetches a job by session id

        Args:
            session_id (str): The session ID

        Returns:
            GetJobResponse: The job fetched
        """
        return await parsed_get_request(
            client=self,
            params={'session_id': session_id},
            url=self._build_endpoint(JobAPIEndpoints.GET_JOBS_USER_FETCH),
            error_message=f"Failed to fetch job for session id {session_id}",
            model=GetJobResponse
        )

    async def list_jobs(self, list_jobs_request: ListJobsRequest) -> ListJobsResponse:
        """
        Lists all jobs for the given user

        Args:
            list_jobs_request (ListJobsRequest): The request including details

        Returns:
            ListJobsResponse: The list of jobs
        """
        return await parsed_post_request(
            client=self,
            json_body=py_to_dict(list_jobs_request),
            params={},
            url=self._build_endpoint(JobAPIEndpoints.POST_JOBS_USER_LIST),
            error_message=f"Failed to list jobs for user.",
            model=ListJobsResponse
        )

    async def list_jobs_in_batch(self, list_request: ListByBatchRequest) -> ListByBatchResponse:
        """
        Gets all jobs within a batch.

        Args:
            list_request (ListByBatchRequest): The request including batch ID

        Returns:
            ListByBatchResponse: The response including list of jobs
        """
        return await parsed_post_request(
            client=self,
            json_body=py_to_dict(list_request),
            params={},
            url=self._build_endpoint(
                JobAPIEndpoints.POST_JOBS_USER_LIST_BATCH),
            error_message=f"Failed to list jobs for user by batch.",
            model=ListByBatchResponse
        )



'''
Created Date: Thursday June 6th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday June 6th 2024 1:39:55 pm +1000
Modified By: Peter Baker
-----
Description: Set of Pydantic models used to define client interfaces which are missing.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from pydantic import BaseModel
from ProvenaInterfaces.RegistryModels import ItemDataset
from typing import List, Union

class SearchItem(BaseModel):
    id: str
    score: float
    
class LoadedSearchItem(SearchItem):
    item: ItemDataset
    
class UnauthorisedSearchItem(SearchItem):
    pass

class FailedSearchItem(SearchItem):
    error_info : str

class RevertMetadata(BaseModel):
    id: str 
    history_id: int
    reason: str

class LoadedSearchResponse(BaseModel):
    # The successfully loaded search results
    items: List[LoadedSearchItem]
    auth_errors: List[UnauthorisedSearchItem]
    misc_errors: List[FailedSearchItem]


'''
Created Date: Friday May 31st 2024 +1000
Author: Peter Baker
-----
Last Modified: Friday May 31st 2024 9:50:26 am +1000
Modified By: Peter Baker
-----
Description: The Auth API L2 Client
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Some issues with admin remove members interfaces - blocked on this
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from ProvenaInterfaces.AuthAPI import *
from provenaclient.clients.client_helpers import *
from provenaclient.models import HealthCheckResponse
from enum import Enum
from provenaclient.utils.helpers import *


class AuthEndpoints(str, Enum):
    # Contains all AuthAPI endpoints - to regen use
    # generate_open_api_endpoint_dump.py with openapi.json as auth-api open api
    # spec (from /openapi.json)

    # NOT IMPLEMENTED
    GET_CHECK_ACCESS_PUBLIC = "/check-access/public"
    GET_CHECK_ACCESS_GENERAL = "/check-access/general"
    GET_ADMIN_CONFIG = "/admin/config"
    GET_ADMIN_SENTRY_DEBUG = "/admin/sentry-debug"

    # DONE
    GET_ACCESS_CONTROL_ADMIN_ALL_PENDING_REQUEST_HISTORY = "/access-control/admin/all-pending-request-history"
    GET_ACCESS_CONTROL_ADMIN_ALL_REQUEST_HISTORY = "/access-control/admin/all-request-history"
    GET_ACCESS_CONTROL_ADMIN_USER_PENDING_REQUEST_HISTORY = "/access-control/admin/user-pending-request-history"
    GET_ACCESS_CONTROL_ADMIN_USER_REQUEST_HISTORY = "/access-control/admin/user-request-history"
    POST_ACCESS_CONTROL_ADMIN_ADD_NOTE = "/access-control/admin/add-note"
    POST_ACCESS_CONTROL_ADMIN_CHANGE_REQUEST_STATE = "/access-control/admin/change-request-state"
    POST_ACCESS_CONTROL_ADMIN_DELETE_REQUEST = "/access-control/admin/delete-request"
    POST_ACCESS_CONTROL_USER_REQUEST_CHANGE = "/access-control/user/request-change"
    GET_ACCESS_CONTROL_USER_REQUEST_HISTORY = "/access-control/user/request-history"
    GET_ACCESS_CONTROL_USER_PENDING_REQUEST_HISTORY = "/access-control/user/pending-request-history"
    GET_ACCESS_CONTROL_USER_GENERATE_ACCESS_REPORT = "/access-control/user/generate-access-report"
    GET_GROUPS_ADMIN_LIST_GROUPS = "/groups/admin/list_groups"
    GET_GROUPS_ADMIN_DESCRIBE_GROUP = "/groups/admin/describe_group"
    GET_GROUPS_ADMIN_LIST_MEMBERS = "/groups/admin/list_members"
    GET_GROUPS_ADMIN_LIST_USER_MEMBERSHIP = "/groups/admin/list_user_membership"
    GET_GROUPS_ADMIN_CHECK_MEMBERSHIP = "/groups/admin/check_membership"
    POST_GROUPS_ADMIN_ADD_MEMBER = "/groups/admin/add_member"
    DELETE_GROUPS_ADMIN_REMOVE_MEMBER = "/groups/admin/remove_member"
    POST_GROUPS_ADMIN_ADD_GROUP = "/groups/admin/add_group"
    DELETE_GROUPS_ADMIN_REMOVE_GROUP = "/groups/admin/remove_group"
    PUT_GROUPS_ADMIN_UPDATE_GROUP = "/groups/admin/update_group"
    GET_GROUPS_ADMIN_EXPORT = "/groups/admin/export"
    POST_GROUPS_ADMIN_IMPORT = "/groups/admin/import"
    POST_GROUPS_ADMIN_RESTORE_FROM_TABLE = "/groups/admin/restore_from_table"
    GET_GROUPS_USER_LIST_GROUPS = "/groups/user/list_groups"
    GET_GROUPS_USER_DESCRIBE_GROUP = "/groups/user/describe_group"
    GET_GROUPS_USER_LIST_USER_MEMBERSHIP = "/groups/user/list_user_membership"
    GET_GROUPS_USER_LIST_MEMBERS = "/groups/user/list_members"
    GET_GROUPS_USER_CHECK_MEMBERSHIP = "/groups/user/check_membership"
    GET_LINK_USER_LOOKUP = "/link/user/lookup"
    POST_LINK_USER_ASSIGN = "/link/user/assign"
    POST_LINK_USER_VALIDATE = "/link/user/validate"
    GET_LINK_ADMIN_LOOKUP = "/link/admin/lookup"
    POST_LINK_ADMIN_ASSIGN = "/link/admin/assign"
    DELETE_LINK_ADMIN_CLEAR = "/link/admin/clear"
    GET_LINK_ADMIN_REVERSE_LOOKUP = "/link/admin/reverse_lookup"
    GET_HEALTH_CHECK = "/"

    # TODO ISSUE
    # Cannot have post body in delete - need to change implementation
    DELETE_GROUPS_ADMIN_REMOVE_MEMBERS = "/groups/admin/remove_members"


class AuthAdminSubClient(ClientService):
    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the AuthClient admin sub client with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_endpoint(self, endpoint: AuthEndpoints) -> str:
        return self._config.auth_api_endpoint + endpoint.value

    async def get_all_pending_request_history(self) -> AccessRequestList:
        """

        Gets all requests with pending status

        Returns:
            AccessRequestList: The response object
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_ADMIN_ALL_PENDING_REQUEST_HISTORY),
            error_message="Failed to retrieve admin pending request history...",
            params={},
            model=AccessRequestList
        )

    async def get_all_request_history(self) -> AccessRequestList:
        """

        Gets all requests

        Returns:
            AccessRequestList: The response object
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_ADMIN_ALL_REQUEST_HISTORY),
            error_message="Failed to retrieve admin request history...",
            params={},
            model=AccessRequestList
        )

    async def get_user_pending_request_history(self, username: str) -> AccessRequestList:
        """

        Gets pending requests for specified user

        Args:
            username (str): Username to query

        Returns:
            AccessRequestList: The response list
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_ADMIN_USER_PENDING_REQUEST_HISTORY),
            error_message="Failed to retrieve user pending request history...",
            params={"username": username},
            model=AccessRequestList
        )

    async def get_user_request_history(self, username: str) -> AccessRequestList:
        """

        Gets all requests for specified user

        Args:
            username (str): Username to query

        Returns:
            AccessRequestList: The response list
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_ADMIN_USER_REQUEST_HISTORY),
            error_message="Failed to retrieve user request history...",
            params={"username": username},
            model=AccessRequestList
        )

    async def post_add_note(self, note: RequestAddNote) -> None:
        """
        Adds a note to an existing request

        Args:
            note (RequestAddNote): Payload incl note info
        """
        await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.POST_ACCESS_CONTROL_ADMIN_ADD_NOTE),
            error_message="Failed to add note.",
            params={},
            json_body=py_to_dict(note),
            model=StatusResponse
        )

    async def post_change_request_state(self, send_email_alert: bool, change: AccessRequestStatusChange) -> ChangeStateStatus:
        """

        Change state of a request. 

        Args:
            send_email_alert (bool): Should trigger email alert?
            change (AccessRequestStatusChange): The details of change

        Returns:
            ChangeStateStatus: The response object
        """
        return await parsed_post_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.POST_ACCESS_CONTROL_ADMIN_CHANGE_REQUEST_STATE),
            error_message="Failed to change request status.",
            params={
                "send_email_alert": send_email_alert
            },
            json_body=py_to_dict(change),
            model=ChangeStateStatus
        )

    async def get_list_groups(self) -> ListGroupsResponse:
        """
        
        Gets a list of groups

        Returns:
            ListGroupsResponse: List of groups
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_ADMIN_LIST_GROUPS),
            error_message="Failed to list admin groups.",
            params={},
            model=ListGroupsResponse
        )

    async def get_describe_group(self, group_id: str) -> DescribeGroupResponse:
        """
        
        Describes a group by ID

        Args:
            group_id (str): The group

        Returns:
            DescribeGroupResponse: Description
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_ADMIN_DESCRIBE_GROUP),
            error_message="Failed to describe group.",
            params={'id': group_id},
            model=DescribeGroupResponse
        )

    async def get_list_members(self, group_id: str) -> ListMembersResponse:
        """
        Lists members of group

        Args:
            group_id (str): The gruop

        Returns:
            ListMembersResponse: The list of members
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_ADMIN_LIST_MEMBERS),
            error_message="Failed to list group members.",
            params={'id': group_id},
            model=ListMembersResponse
        )

    async def get_list_group_membership(self, username: str) -> ListUserMembershipResponse:
        """
        
        Gets list of groups a user is in

        Args:
            username (str): The username

        Returns:
            ListUserMembershipResponse: The list of groups
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_ADMIN_LIST_MEMBERS),
            error_message="Failed to list user group membership.",
            params={'username': username},
            model=ListUserMembershipResponse
        )

    async def get_check_user_membership(self, username: str, group_id: str) -> CheckMembershipResponse:
        """
        
        Checks user membership within a group

        Args:
            username (str): The username to target
            group_id (str): The group to check

        Returns:
            CheckMembershipResponse: Response
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_ADMIN_CHECK_MEMBERSHIP),
            error_message="Failed to list user group membership.",
            params={'username': username, 'group_id': group_id},
            model=CheckMembershipResponse
        )

    async def post_groups_add_member(self, group_id: str, user: GroupUser) -> AddMemberResponse:
        """
        
        Adds a member to a group

        Args:
            group_id (str): Id of group
            user (GroupUser): The user to add

        Returns:
            AddMemberResponse: The response
        """
        return await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.POST_GROUPS_ADMIN_ADD_MEMBER),
            error_message="Failed to add user to group",
            params={'group_id': group_id},
            json_body=py_to_dict(user),
            model=AddMemberResponse
        )

    async def delete_remove_member(self, group_id: str, username: str) -> RemoveMemberResponse:
        """
        
        Removes member from group

        Args:
            group_id (str): The id of group
            username (str): The user to remove

        Returns:
            RemoveMemberResponse: The response
        """
        return await parsed_delete_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.DELETE_GROUPS_ADMIN_REMOVE_MEMBER),
            error_message="Failed to remove member from group.",
            params={'group_id': group_id, 'username': username},
            model=RemoveMemberResponse
        )

    async def post_add_group(self, group: UserGroupMetadata) -> AddGroupResponse:
        """
        
        Adds a group/creates group

        Args:
            group (UserGroupMetadata): The group details

        Returns:
            AddGroupResponse: The response
        """
        return await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.POST_GROUPS_ADMIN_ADD_GROUP),
            error_message="Failed to add group.",
            json_body=py_to_dict(group),
            params={},
            model=AddGroupResponse
        )

    async def put_update_group(self, group: UserGroupMetadata) -> UpdateGroupResponse:
        """
        Updates group details

        Args:
            group (UserGroupMetadata): The group metadata

        Returns:
            UpdateGroupResponse: The response
        """
        return await parsed_put_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.PUT_GROUPS_ADMIN_UPDATE_GROUP),
            error_message="Failed to update gruop.",
            json_body=py_to_dict(group),
            params={},
            model=UpdateGroupResponse
        )

    async def get_export_groups(self) -> GroupsExportResponse:
        """
        Exports all group details in specified format.

        Returns:
            GroupsExportResponse: The data dump
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(AuthEndpoints.GET_GROUPS_ADMIN_EXPORT),
            error_message="Failed to export groups.",
            params={},
            model=GroupsExportResponse
        )

    async def post_import_groups(self, body: GroupsImportRequest) -> GroupsImportResponse:
        """
        
        Imports groups from data dump back in

        Args:
            body (GroupsImportRequest): The import request incl. dump

        Returns:
            GroupsImportResponse: The response
        """
        return await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(AuthEndpoints.POST_GROUPS_ADMIN_IMPORT),
            error_message="Groups import failed.",
            json_body=py_to_dict(body),
            params={},
            model=GroupsImportResponse
        )

    async def post_restore_groups_from_table(self, table_name: str, body: GroupsRestoreRequest) -> GroupsImportResponse:
        """
        
        Restores groups by first dumping from valid group table. 
        Needs permissions to table.

        Args:
            table_name (str): The table name
            body (GroupsRestoreRequest): The request

        Returns:
            GroupsImportResponse: The response details
        """
        return await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.POST_GROUPS_ADMIN_RESTORE_FROM_TABLE),
            error_message="Groups restore from table failed.",
            json_body=py_to_dict(body),
            params={'table_name': table_name},
            model=GroupsImportResponse
        )

    async def get_link_lookup_username(self, username: str) -> AdminLinkUserLookupResponse:
        """
        Gets the linked person for user

        Args:
            username (str): User to lookup

        Returns:
            AdminLinkUserLookupResponse: Response
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(AuthEndpoints.GET_LINK_ADMIN_LOOKUP),
            error_message="Failed to lookup user in link service.",
            params={'username': username},
            model=AdminLinkUserLookupResponse
        )

    async def get_link_reverse_lookup_username(self, person_id: str) -> UserLinkReverseLookupResponse:
        """
        
        Looks up reverse by person ID

        Args:
            person_id (str): The person to lookup

        Returns:
            UserLinkReverseLookupResponse: The response
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_LINK_ADMIN_REVERSE_LOOKUP),
            error_message="Failed to lookup reverse in link service.",
            params={'person_id': person_id},
            model=UserLinkReverseLookupResponse
        )

    async def post_link_assign(self, body: AdminLinkUserAssignRequest) -> AdminLinkUserAssignResponse:
        """
        
        Assigns a person to a given user

        Args:
            body (AdminLinkUserAssignRequest): The request

        Returns:
            AdminLinkUserAssignResponse: The response
        """
        return await parsed_post_request(
            client=self,
            url=self._build_endpoint(AuthEndpoints.POST_LINK_ADMIN_ASSIGN),
            error_message="Failed to assign user link.",
            params={},
            json_body=py_to_dict(body),
            model=AdminLinkUserAssignResponse
        )

    async def delete_clear_link(self, username: str) -> AdminLinkUserClearResponse:
        """
        Deletes an existing link

        Args:
            username (str): The user to unlink

        Returns:
            AdminLinkUserClearResponse: The response
        """
        return await parsed_delete_request_with_status(
            client=self,
            url=self._build_endpoint(AuthEndpoints.DELETE_LINK_ADMIN_CLEAR),
            error_message="Failed to clear link from user.",
            params={'username': username},
            model=AdminLinkUserClearResponse
        )

    # TODO this is not currently possible because the DELETE method should not have JSON body according to HTTPX
    # async def post_remove_members(self, group_id: str, username: str) -> RemoveMemberResponse:
    #    return await parsed_delete_request_with_status(
    #        client=self,
    #        url=self._build_endpoint(AuthEndpoints.DELETE_GROUPS_ADMIN_REMOVE_MEMBERS),
    #        error_message="Failed to remove member from group.",
    #        params={'group_id': group_id, 'username': username},
    #        model=RemoveMemberResponse
    #    )

# L2 interface.


class AuthClient(ClientService):
    # Sub clients
    admin: AuthAdminSubClient

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the AuthClient with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

        self.admin = AuthAdminSubClient(auth=auth, config=config)

    def _build_endpoint(self, endpoint: AuthEndpoints) -> str:
        return self._config.auth_api_endpoint + endpoint.value

    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns:
            HealthCheckResponse: Response
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(AuthEndpoints.GET_HEALTH_CHECK),
            error_message="Health check failed!",
            params={},
            model=HealthCheckResponse
        )

    async def get_user_request_history(self) -> AccessRequestList:
        """
        
        Gets the users access request history

        Returns:
            AccessRequestList: The list of requests
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_USER_REQUEST_HISTORY),
            error_message="Failed to retrieve access request history.",
            params={},
            model=AccessRequestList
        )

    async def post_user_request_change(self, body: AccessReport, send_email: bool) -> StatusResponse:
        """
        
        Requests a change by diffing access models

        Args:
            body (AccessReport): The new access desired
            send_email (bool): Email alert

        Returns:
            StatusResponse: Ok?
        """
        return await parsed_post_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.POST_ACCESS_CONTROL_USER_REQUEST_CHANGE),
            error_message="Failed to request access change.",
            params={"send_email": send_email},
            json_body=py_to_dict(body),
            model=StatusResponse
        )

    async def get_user_pending_request_history(self) -> AccessRequestList:
        """
        Gets only pending requests from history

        Returns:
            AccessRequestList: The list
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_USER_PENDING_REQUEST_HISTORY),
            error_message="Failed to retrieve access pending request history.",
            params={},
            model=AccessRequestList
        )

    async def get_user_generate_access_report(self) -> AccessReportResponse:
        """
        
        Generates an access report detailing system access.

        Returns:
            AccessReportResponse: The response
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_ACCESS_CONTROL_USER_GENERATE_ACCESS_REPORT),
            error_message="Failed to generate access report.",
            params={},
            model=AccessReportResponse
        )

    async def get_list_groups(self) -> ListGroupsResponse:
        """
        Lists all groups

        Returns:
            ListGroupsResponse: List of groups
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_USER_LIST_GROUPS),
            error_message="Failed to list groups",
            params={},
            model=ListGroupsResponse
        )

    async def get_describe_group(self, group_id: str) -> DescribeGroupResponse:
        """
        
        Describes a specific gruop

        Args:
            group_id (str): The id of group

        Returns:
            DescribeGroupResponse: Response with details
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_USER_DESCRIBE_GROUP),
            error_message="Failed to describe group",
            params={'id': group_id},
            model=DescribeGroupResponse
        )

    async def get_list_membership(self) -> ListUserMembershipResponse:
        """
        
        Gets the list of groups user is member of

        Returns:
            ListUserMembershipResponse: List and details
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_USER_LIST_USER_MEMBERSHIP),
            error_message="Failed to list group membership.",
            params={},
            model=ListUserMembershipResponse
        )

    async def get_list_group_members(self, group_id: str) -> ListMembersResponse:
        """
        
        Lists the members of a given group

        Args:
            group_id (str): The group to lookup

        Returns:
            ListMembersResponse: Members
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_USER_LIST_MEMBERS),
            error_message="Failed to list group members",
            params={'id': group_id},
            model=ListMembersResponse
        )

    async def get_check_membership(self, group_id: str) -> CheckMembershipResponse:
        """
        Checks if user is in specific group

        Args:
            group_id (str): The group to check

        Returns:
            CheckMembershipResponse: In group?
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                AuthEndpoints.GET_GROUPS_USER_CHECK_MEMBERSHIP),
            error_message="Failed to check membership.",
            params={'id': group_id},
            model=CheckMembershipResponse
        )

    async def get_link_lookup_username(self, username: Optional[str]=None) -> UserLinkUserLookupResponse:
        """
        
        Looks up either current user or specified user

        Args:
            username (Optional[str], optional): The username if not current user. Defaults to None.

        Returns:
            UserLinkUserLookupResponse: The response indicating link
        """
        return await parsed_get_request(
            client=self,
            url=self._build_endpoint(AuthEndpoints.GET_LINK_USER_LOOKUP),
            error_message="Failed to lookup user in link service.",
            params={'username': username},
            model=UserLinkUserLookupResponse
        )

    async def post_link_assign(self, body: UserLinkUserAssignRequest) -> UserLinkUserAssignResponse:
        """
        
        Assigns link to current user.

        Args:
            body (UserLinkUserAssignRequest): The link to assign

        Returns:
            UserLinkUserAssignResponse: The response indicating success
        """
        return await parsed_post_request(
            client=self,
            url=self._build_endpoint(AuthEndpoints.POST_LINK_USER_ASSIGN),
            error_message="Failed to assign user link.",
            params={},
            json_body=py_to_dict(body),
            model=UserLinkUserAssignResponse
        )

    async def post_link_validate(self, body: UserLinkUserAssignRequest) -> UserLinkUserValidateResponse:
        """
        Validates link before making it.

        Args:
            body (UserLinkUserAssignRequest): The link to assign

        Returns:
            UserLinkUserValidateResponse: Valid?
        """
        # Doesn't assert status as we want the status response to fall through
        # to user
        return await parsed_post_request(
            client=self,
            url=self._build_endpoint(AuthEndpoints.POST_LINK_USER_VALIDATE),
            error_message="Failed to validate user link.",
            params={},
            json_body=py_to_dict(body),
            model=UserLinkUserValidateResponse
        )



'''
Created Date: Thursday May 30th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday May 30th 2024 10:17:36 am +1000
Modified By: Peter Baker
-----
Description: Search API L2 Client.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from enum import Enum
from ProvenaInterfaces.RegistryModels import ItemSubType
from ProvenaInterfaces.SearchAPI import QueryResults
from provenaclient.utils.helpers import *
from provenaclient.clients.client_helpers import *


class SearchEndpoints(str, Enum):
    """An ENUM containing the datastore-api
    endpoints.
    """
    SEARCH_REGISTRY: str = "/search/entity-registry"

# L2 interface.
class SearchClient(ClientService):
    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the SearchClient with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_endpoint(self, endpoint: SearchEndpoints) -> str:
        return self._config.search_api_endpoint + endpoint.value

    async def search_registry(self, query: str, limit: Optional[int], subtype_filter: Optional[ItemSubType]) -> QueryResults:
        """

        Searches registry using search API for given query, limit and subtype.

        Args:
            query (str): The query
            limit (Optional[int]): The record result limit
            subtype_filter (Optional[ItemSubType]): The subtype to filter if desired

        Raises:
            e: Generic exceptions as usual

        Returns:
            QueryResults: The results, not loaded.
        """
        return await parsed_get_request_with_status(
            client=self,
            url=self._build_endpoint(
                SearchEndpoints.SEARCH_REGISTRY),
            error_message=f"Search with query '{query}' failed!...",
            params={"query": query, "record_limit": limit, "subtype_filter": subtype_filter.value if subtype_filter else None},
            model=QueryResults
        )


# read version from installed package
from importlib.metadata import version
__version__ = version("provenaclient")

from provenaclient.modules.provena_client import ProvenaClient
from provenaclient.utils.config import Config, EndpointConfig

import ProvenaInterfaces as interfaces


'''
Created Date: Thursday June 6th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday June 6th 2024 1:39:55 pm +1000
Modified By: Peter Baker
-----
Description: Datastore L3 module. Includes the Data store review sub module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import DatastoreClient, SearchClient
from ProvenaInterfaces.DataStoreAPI import *
from ProvenaInterfaces.RegistryModels import CollectionFormat, ItemSubType
from provenaclient.models import HealthCheckResponse, LoadedSearchResponse, LoadedSearchItem, UnauthorisedSearchItem, FailedSearchItem, RevertMetadata
from provenaclient.utils.exceptions import *
from provenaclient.modules.module_helpers import *
from ProvenaInterfaces.RegistryAPI import NoFilterSubtypeListRequest, VersionRequest, VersionResponse, SortOptions, SortType, DatasetListResponse
from provenaclient.modules.submodules import IOSubModule

from typing import AsyncGenerator, List

# L3 interface.

DEFAULT_SEARCH_LIMIT = 25
DATASTORE_DEFAULT_SEARCH_LIMIT = 20


class ReviewSubModule(ModuleService):
    _datastore_client: DatastoreClient

    def __init__(self, auth: AuthManager, config: Config, datastore_client: DatastoreClient) -> None:
        """
        System reviewer sub module of the Datastore API functionality

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        auth_client: AuthClient
            The instantiated auth client
        """
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._datastore_client = datastore_client

    async def delete_dataset_reviewer(self, reviewer_id: str) -> None:
        """Delete a reviewer.  

        Parameters
        ----------
        reviewer_id : str
            Id of an existing reviewer within the system.
        """
        await self._datastore_client.review.delete_dataset_reviewer(reviewer_id=reviewer_id)

    async def add_dataset_reviewer(self, reviewer_id: str) -> None:
        """Add a reviewer.

        Parameters
        ----------
        reviewer_id : str
            Id of a reviewer.

        """

        await self._datastore_client.review.add_dataset_reviewer(reviewer_id=reviewer_id)

    """
    async def list_reviewers(self) -> TODO: 
        TODO

    Leaving this for now,as the pydantic response model of this endpoint will be duly updated.
    """

    async def dataset_approval_request(self, approval_request: ReleaseApprovalRequest) -> ReleaseApprovalRequestResponse:
        """Submit a request for approval of dataset.

        Parameters
        ----------
        approval_request : ReleaseApprovalRequest
            An object that requires the dataset id, approver id and notes

        Returns
        -------
        ReleaseApprovalRequestResponse:
            Contains details of the approval request.

        """

        return await self._datastore_client.review.approval_request(approval_request_payload=approval_request)

    async def action_approval_request(self, action_approval_request: ActionApprovalRequest) -> ActionApprovalRequestResponse:
        """Action an approval request from a dataset approval request via the datastore.

        Parameters
        ----------
        action_approval_request : ActionApprovalRequest
            The dataset id, your decision of approval and any extra information 
            you want to add (notes).

        Returns
        -------
        ActionApprovalRequestResponse
            The details of the approval action and the relevant dataset details.
        """

        return await self._datastore_client.review.action_approval_request(action_approval_request_payload=action_approval_request)


class Datastore(ModuleService):
    _datastore_client: DatastoreClient
    _search_client: SearchClient

    # sub modules
    review: ReviewSubModule
    io: IOSubModule

    def __init__(self, auth: AuthManager, config: Config, datastore_client: DatastoreClient, search_client: SearchClient) -> None:
        """Initialise a new datastore object, which sits between the user and the datastore api operations.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance. 
        datastore_client : DatastoreClient
            This client interacts with the Datastore API's.
        """
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._datastore_client = datastore_client
        self._search_client = search_client
        self.review = ReviewSubModule(
            auth=auth,
            config=config,
            datastore_client=self._datastore_client
        )
        self.io = IOSubModule(
            auth=auth,
            config=config,
            datastore_client=self._datastore_client
        )

    async def get_health_check(self) -> HealthCheckResponse:
        """Health check the API

        Returns
        -------
        HealthCheckResponse
            Response
        """

        return await self._datastore_client.get_health_check()

    async def fetch_dataset(self, id: str) -> RegistryFetchResponse:
        """Fetches a dataset from the datastore based on the provided
        ID.


        Parameters
        ----------
        id : str
            The unique identifier of the dataset to be retrieved.
            For example: "10378.1/1451860"

        Returns
        -------
        RegistryFetchResponse
            A interactive python datatype of type RegistryFetchResponse
            containing the dataset details.

        """

        return await self._datastore_client.fetch_dataset(id)

    async def mint_dataset(self, dataset_mint_info: CollectionFormat) -> MintResponse:
        """Creates a new dataset in the datastore with the provided dataset information.

        Parameters
        ----------
        dataset_mint_info : CollectionFormat
            A structured format containing all necessary information to register a new dataset, including associations, 
            approvals, and dataset-specific information.

        Returns
        -------
        MintResponse
            A interactive python datatype of type MintResponse
            containing the newly created dataset details.

        """

        return await self._datastore_client.mint_dataset(dataset_mint_info)

    async def validate_dataset_metadata(self, metadata_payload: CollectionFormat) -> StatusResponse:
        """Validates the dataset metadata creation for testing and does not publish.

        Parameters
        ----------
        metadata_payload : CollectionFormat
            A structured format containing all necessary information to register a new dataset, including associations, 
            approvals, and dataset-specific information.

        Returns
        -------
        StatusResponse
            Response indicating whether your dataset metadata setup is valid and correct.
        """

        return await self._datastore_client.validate_metadata(metadata_payload=metadata_payload)

    async def update_dataset_metadata(self, handle_id: str, reason: str, metadata_payload: CollectionFormat) -> UpdateMetadataResponse:
        """Updates an existing dataset's metadata.

        Parameters
        ----------
        handle_id : str
            The id of the dataset.
        reason : str
            The reason for changing metadata of the dataset.
        metadata_payload : CollectionFormat
            A structured format containing all necessary information to register a new dataset, including associations, 
            approvals, and dataset-specific information.

        Returns
        -------
        UpdateMetadataResponse
            The updated metadata response
        """

        return await self._datastore_client.update_metadata(handle_id=handle_id, reason=reason, metadata_payload=metadata_payload)

    async def revert_dataset_metadata(self, metadata_payload: RevertMetadata) -> StatusResponse:
        """Reverts the metadata for a dataset to a previous identified historical version.

        Parameters
        ----------
        metadata_payload : RevertMetadata
            The revert request, passed through to the registry API and requires
            dataset id, history id and reason for reverting.

        Returns
        -------
        StatusResponse
            Response indicating whether your dataset revert metadata request was correct.
        """

        return await self._datastore_client.revert_metadata(metadata_payload=metadata_payload)

    async def version_dataset(self, version_request: VersionRequest) -> VersionResponse:
        """Versioning operation which creates a new version from the specified ID.

        Parameters
        ----------
        version_request : VersionRequest
            The request which includes the item ID and reason for versioning.

        Returns
        -------
        VersionResponse
            Response of the versioning of the dataset, containing new version ID and 
            job session ID.
        """

        return await self._datastore_client.version_dataset(version_dataset_payload=version_request)

    async def for_all_datasets(self, list_dataset_request: NoFilterSubtypeListRequest, total_limit: Optional[int] = None) -> AsyncGenerator[ItemDataset, None]:
        """Fetches all datasets based on the provided datasets in datastore based on 
            the provided sorting criteria, pagination key and page size. 

        Parameters
        ----------
        list_dataset_request : NoFilterSubtypeListRequest
            A request object configured with sorting options, pagination keys, 
            and page size that defines how datasets are queried from the datastore. 
        total_limit : Optional[int], optional
            A maximum number of datasets to fetch. If specified, the generator will 
            stop yielding datasets once this limit is reached. 
            If None, it will fetch datasets until there are no more to fetch.

        Returns
        -------
        AsyncGenerator[ItemDataset, None]
            An asynchronous generator yielding "ItemDataset"
            object which is an individual dataset from the datastore.

        Yields
        ------
        Iterator[AsyncGenerator[ItemDataset, None]]
            Each yield provides a "ItemDataset" containing an individual dataset.

        """

        total_fetched = 0

        while True:
            response = await self._datastore_client.list_datasets(list_request=list_dataset_request)

            # We will keep track of the total numbers of items fetched in the request from the API itself.
            if response.items:
                for dataset in response.items:
                    if total_limit is not None and total_fetched >= total_limit:
                        return
                    yield dataset
                    total_fetched = total_fetched + 1
            else:
                # If there are issues with gathering responses, we will increment based on the page size.
                total_fetched = total_fetched + list_dataset_request.page_size

            if response.pagination_key is None:
                break

            # Update the pagination key for the next set of requests.
            list_dataset_request.pagination_key = response.pagination_key

    async def list_datasets(self, list_dataset_request: NoFilterSubtypeListRequest) -> DatasetListResponse:
        """Takes a specific dataset list request and returns the response.

        Parameters
        ----------
        list_dataset_request : NoFilterSubtypeListRequest
            A request object configured with sorting options, pagination keys, 
            and page size that defines how datasets are queried from the datastore. 

        Returns
        -------
        DatasetListResponse
            Response containing the requested datasets in the datastore 
            based on sort criteria and page size, and contains other attributes
            such as total_item_counts and optional pagination key.
        """

        datasets = await self._datastore_client.list_datasets(list_request=list_dataset_request)
        return datasets

    async def list_all_datasets(self, sort_criteria: Optional[SortOptions] = None) -> List[ItemDataset]:
        """Fetches all datasets from the datastore and you may provide your own sort criteria.
        By default uses display name sort criteria. 

        Parameters
        ----------
        sort_criteria : Optional[SortOptions]
            An object configured with sorting options that you want 
            when displaying all datasets within the datastore.

        Returns
        -------
        List[ItemDataset]
            A list of all datasets in the datastore, sorted as requested.
        """

        list_dataset_request: NoFilterSubtypeListRequest = NoFilterSubtypeListRequest(
            sort_by=sort_criteria,
            pagination_key=None,
            page_size=DATASTORE_DEFAULT_SEARCH_LIMIT
        )

        combined_dataset_list: List[ItemDataset] = []

        while True:
            response = await self._datastore_client.list_datasets(list_request=list_dataset_request)

            if response.items:
                combined_dataset_list.extend(response.items)

            if response.pagination_key is None:
                break

            # Update the pagination key for the next request
            list_dataset_request.pagination_key = response.pagination_key

        return combined_dataset_list

    async def generate_dataset_presigned_url(self, dataset_presigned_request: PresignedURLRequest) -> PresignedURLResponse:
        """Generates a presigned url for an existing dataset.

        Parameters
        ----------
        dataset_presigned_request : PresignedURLRequest
            Contains the dataset id + file path + length of expiry of URL.

        Returns
        -------
        PresignedURLResponse
            A response with the presigned url.
        """

        return await self._datastore_client.generate_presigned_url(presigned_url=dataset_presigned_request)

    async def generate_read_access_credentials(self, credentials: CredentialsRequest) -> CredentialResponse:
        """Given an S3 location, will attempt to generate programmatic access keys for 
           the storage bucket at this particular subdirectory.

        Parameters
        ----------
        credentials : CredentialsRequest
            Contains the dataset id + console session URL required flag (boolean)

        Returns
        -------
        CredentialResponse
            The AWS credentials creating read level access into the subset of the bucket requested in the S3 location object.
        """

        return await self._datastore_client.generate_read_access_credentials(read_access_credentials=credentials)

    async def generate_write_access_credentials(self, credentials: CredentialsRequest) -> CredentialResponse:
        """Given an S3 location, will attempt to generate programmatic access keys for 
           the storage bucket at this particular subdirectory.

        Parameters
        ----------
        credentials : CredentialsRequest
            Contains the dataset id + console session URL required flag (boolean)

        Returns
        -------
        CredentialResponse
            The AWS credentials creating write level access into the subset of the bucket requested in the S3 location object.
        """

        return await self._datastore_client.generate_write_access_credentials(write_access_credentials=credentials)

    async def search_datasets(self, query: str, limit: int = DEFAULT_SEARCH_LIMIT) -> LoadedSearchResponse:
        """

        Utilises the L2 search client to search for datasets with the specified
        query.

        Loads all datasets in the result payload from the data store and sorts
        based on auth, or other exceptions if not successful.

        Args:
            query (str): The query to make limit (int, optional): The result
            count limit. Defaults to DEFAULT_SEARCH_LIMIT.

        Returns:
            LoadedSearchResponse: The loaded items incl errors.
        """
        # search with search client
        search_results = await self._search_client.search_registry(
            query=query,
            limit=limit,
            subtype_filter=ItemSubType.DATASET
        )

        # load each result checking 401 for auth errors or misc errors recorded
        success: List[LoadedSearchItem] = []
        auth_err: List[UnauthorisedSearchItem] = []
        misc_err: List[FailedSearchItem] = []

        assert search_results.results is not None
        for item in search_results.results:
            # load item
            try:
                loaded_dataset = await self._datastore_client.fetch_dataset(
                    id=item.id
                )
                assert loaded_dataset.item
                success.append(LoadedSearchItem(
                    id=item.id,
                    item=loaded_dataset.item,
                    score=item.score
                ))
            except AuthException as e:
                auth_err.append(UnauthorisedSearchItem(
                    id=item.id,
                    score=item.score
                ))
            except Exception as e:
                misc_err.append(FailedSearchItem(
                    id=item.id,
                    score=item.score,
                    error_info=f"Failed to fetch item, error: {e}."
                ))

        return LoadedSearchResponse(
            items=success,
            auth_errors=auth_err,
            misc_errors=misc_err
        )



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Set of helpers for running polling tasks against the job API until completion etc.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from typing import Dict, Any, Callable, Optional, cast, Tuple, Coroutine
from ProvenaInterfaces.AsyncJobModels import JobStatusTable
from ProvenaInterfaces.AsyncJobAPI import *
from datetime import datetime
from time import sleep
from provenaclient.clients import JobAPIClient
from provenaclient.models import AsyncAwaitSettings
from provenaclient.utils.exceptions import BaseException

JOB_FINISHED_STATES = [JobStatus.FAILED, JobStatus.SUCCEEDED]

Payload = Dict[str, Any]


def timestamp() -> int:
    return int(datetime.now().timestamp())


# finished? optional data
PollCallbackData = Optional[Any]
PollCallbackResponse = Tuple[bool, PollCallbackData]
PollCallbackFunction = Callable[[],
                                Coroutine[Any, Any, PollCallbackResponse]]


class PollTimeoutException(Exception):
    "Raised when poll operation exceeds timeout"
    pass


class PollFunctionErrorException(Exception):
    "Raised when poll operation callback errors"
    pass


async def poll_callback(poll_interval_seconds: int, timeout_seconds: int, callback: PollCallbackFunction) -> PollCallbackData:
    """
    The poll callback function which repeatedly polls given func according to spec

    Args:
        poll_interval_seconds (int): The polling interval
        timeout_seconds (int): The timeout to adhere to
        callback (PollCallbackFunction): The call back function itself

    Raises:
        PollFunctionErrorException: Unexpected error
        PollTimeoutException: Timeout

    Returns:
        PollCallbackData: The resulting data (job status table probably)
    """
    start_time = timestamp()
    completed = False
    data = None

    def timeout_condition_check() -> bool:
        curr = timestamp()
        return curr - start_time <= timeout_seconds

    def print_status() -> None:
        print(
            f"Polling Job API. Wait time: {round(timestamp() - start_time,2)}sec out of {timeout_seconds}sec.")

    while (not completed) and timeout_condition_check():
        print_status()

        try:
            fin, res_data = await callback()
        except Exception as e:
            raise PollFunctionErrorException(
                f"Polling callback function raised an error. Aborting. Error: {e}.")

        if fin:
            data = res_data
            completed = True
        else:
            print("Callback registered incomplete. Waiting for polling interval.")
            sleep(poll_interval_seconds)

    if completed:
        return data
    else:
        raise PollTimeoutException("Timed out during polling operation.")


async def wait_for_in_progress(session_id: str, client: JobAPIClient, settings: AsyncAwaitSettings) -> JobStatusTable:
    """
    Waits for the entry to be in in progress status

    Args:
        session_id (str): The session ID
        client (JobAPIClient): The L2 job client
        settings (AsyncAwaitSettings): The settings

    Returns:
        JobStatusTable: The resulting entry
    """
    timeout = settings.job_async_pending_polling_timeout
    interval = settings.job_polling_interval

    async def callback() -> PollCallbackResponse:
        print(
            f"Running wait for in progress callback. Session ID: {session_id}.")
        entry = (await client.fetch_job(session_id=session_id)).job

        # stop if we exit pending - could jump straight to error/success otherwise
        if entry.status != JobStatus.PENDING:
            print(
                f"200OK response for user fetch of {session_id} in non pending state {entry.status}.")
            return True, entry
        else:
            print(
                f"200OK response for user fetch of {session_id} in state {entry.status}.")
            return False, None

    # poll
    print(f"Starting wait_for_in_progress polling stage.")
    data = await poll_callback(
        poll_interval_seconds=interval,
        timeout_seconds=timeout,
        callback=callback
    )
    print(f"Finished wait_for_in_progress polling stage.")

    assert data is not None
    return cast(JobStatusTable, data)


async def wait_for_entry_in_queue(session_id: str, client: JobAPIClient, settings: AsyncAwaitSettings) -> JobStatusTable:
    """
    Waits for the entry to appear in the queue

    Args:
        session_id (str): The session ID
        client (JobAPIClient): The L2 job client
        settings (AsyncAwaitSettings): The settings

    Raises:
        Exception: If non 400 error

    Returns:
        JobStatusTable: The resulting entry
    """
    timeout = settings.job_async_queue_delay_polling_timeout
    interval = settings.job_polling_interval

    async def callback() -> PollCallbackResponse:
        print(
            f"Running wait_for_entry_in_queue callback. Session ID: {session_id}.")
        try:
            job = await client.fetch_job(session_id=session_id)
        except BaseException as be:
            if be.error_code != 400:
                raise Exception(
                    f"Unexpected error state when waiting for job. Code: {be.error_code}. Error: {be.message}.") from be
            else:
                return False, None

        print(f"200OK response for user fetch of {session_id}.")
        # Parse and return the 200OK data
        entry: JobStatusTable = job.job
        return True, entry

    # poll
    print(f"Starting wait_for_entry_in_queue polling stage.")
    data = await poll_callback(
        poll_interval_seconds=interval,
        timeout_seconds=timeout,
        callback=callback
    )
    print(f"Finished wait_for_entry_in_queue polling stage.")

    return cast(JobStatusTable, data)


async def wait_for_completion(session_id: str, client: JobAPIClient, settings: AsyncAwaitSettings) -> JobStatusTable:
    """
    Waits for completion.

    Args:
        session_id (str): The session ID
        client (JobAPIClient): The L2 job client
        settings (AsyncAwaitSettings): The settings

    Returns:
        JobStatusTable: The resulting entry
    """
    timeout = settings.job_async_in_progress_polling_timeout
    interval = settings.job_polling_interval

    async def callback() -> PollCallbackResponse:
        print(
            f"Running wait for completion callback. Session ID: {session_id}.")

        data = await client.fetch_job(session_id=session_id)
        entry = data.job

        # stop if we exit pending - could jump straight to error/success otherwise
        if entry.status in JOB_FINISHED_STATES:
            print(
                f"200OK response for user fetch of {session_id} in completed state.")
            return True, entry
        else:
            print(
                f"200OK response for user fetch of {session_id} in state {entry}.")
            return False, None

    # poll
    print(f"Starting wait_for_completion polling stage.")
    data = await poll_callback(
        poll_interval_seconds=interval,
        timeout_seconds=timeout,
        callback=callback
    )
    print(f"Finished wait_for_completion polling stage.")

    assert data is not None
    return cast(JobStatusTable, data)


async def wait_for_full_lifecycle(session_id: str, client: JobAPIClient, settings: AsyncAwaitSettings) -> JobStatusTable:
    """

    Waits for the entire lifecycle of a job, adhering to settings. Uses L2 client.

    Args:
        session_id (str): The session ID to await
        client (JobAPIClient): The client to utilise
        settings (AsyncAwaitSettings): The settings including timeouts

    Returns:
        JobStatusTable: The resulting entry
    """
    # wait for item to be present in job table - could take a few seconds
    data = await wait_for_entry_in_queue(session_id=session_id, client=client, settings=settings)

    # backout early
    if data.status in JOB_FINISHED_STATES:
        return data

    # block into two steps to time separately
    data = await wait_for_in_progress(session_id=session_id, client=client, settings=settings)

    # backout early
    if data.status in JOB_FINISHED_STATES:
        return data

    return await wait_for_completion(session_id=session_id, client=client, settings=settings)


async def wait_for_full_successful_lifecycle(session_id: str,  client: JobAPIClient, settings: AsyncAwaitSettings) -> JobStatusTable:
    """
    Waits for the full lifecycle of an async job and asserts that it is successful at the end.

    Args:
        session_id (str): The session ID to await
        client (JobAPIClient): The client 
        settings (AsyncAwaitSettings): The settings

    Returns:
        JobStatusTable: The resulting successful job
    """
    res = await wait_for_full_lifecycle(session_id=session_id, client=client, settings=settings)

    assert res.status == JobStatus.SUCCEEDED, \
        f"Job failed, error {res.info or 'None provided'}. Session ID {session_id}."

    assert res.result is not None, \
        f"Job succeeded, but did not include a result payload!"

    return res



'''
Created Date: Tuesday June 18th 2024 +1000
Author: peter
-----
Last Modified: Tuesday June 18th 2024 12:44:03 pm +1000
Modified By: peter
-----
Description: Miscellaneous helper functions assisting with implementation of auth flows
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from pydantic import BaseModel
from typing import Optional, Any
from datetime import datetime, timezone
from typing import Generator, Optional, Dict, List
from httpx import Auth, Request, Response
from jose import jwt, JWTError # type: ignore
from jose.constants import ALGORITHMS # type: ignore
import requests
import logging


# Default constant value for JWT Expiry Window in seconds.
JWT_DEFAULT_WINDOW = 30


class AccessToken(BaseModel):
    access_token: str


class Tokens(BaseModel):
    access_token: str
    # refresh tokens are marked as optional because offline tokens should not be cached
    refresh_token: Optional[str]


class HttpxBearerAuth(Auth):
    def __init__(self, token: str) -> None:
        self.token = token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers['Authorization'] = "Bearer " + self.token
        yield request


def check_token_expiry_window(jwt_data: dict[str, Any], logger: logging.Logger, jwt_token_expiry_window: int = JWT_DEFAULT_WINDOW) -> bool:
    """This helper function checks if the current JWT token will 
    expire or not expire either within the provided or default (30sec)  window.
    If the token is going to be expired in less than or within the provided or default (30sec).
    expiry window they will be refreshed. 

    Parameters
    ----------
    jwt_data : dict[str,Any]
        A dictionary containing the token validation results.
    jwt_token_expiry_window: Optional[int]
        A potential integer value containing your desired JWT expiry 
        window.

    Returns
    -------
    bool
        True: The current token will not expire within 30 seconds
        False: The current token will expire within 30 seconds.
    """

    # Contains an unix timestamp
    expiration_timestamp = jwt_data.get("exp")

    if expiration_timestamp:

        # We will need to convert to a datetime/utc object here.
        expiration_time = datetime.fromtimestamp(
            expiration_timestamp, timezone.utc)
        current_time = datetime.now(timezone.utc)
        remaining_time = (expiration_time - current_time).total_seconds()

        if remaining_time <= jwt_token_expiry_window:
            return False

        else:
            return True

    return False

def retrieve_keycloak_public_key(keycloak_endpoint: str, logger: logging.Logger) -> str:
        """Given the keycloak endpoint, retrieves the advertised
        public key.
        Based on https://github.com/nurgasemetey/fastapi-keycloak-oidc/blob/main/main.py
        """
        error_message = f"Error finding public key from keycloak endpoint {keycloak_endpoint}."
        try:
            r = requests.get(url=keycloak_endpoint,
                             timeout=3)
            r.raise_for_status()
            response_json = r.json()
            return f"-----BEGIN PUBLIC KEY-----\r\n{response_json['public_key']}\r\n-----END PUBLIC KEY-----"
        except requests.exceptions.HTTPError as errh:
            logger.info(error_message)
            logger.info("Http Error:" + str(errh))
            raise errh
        except requests.exceptions.ConnectionError as errc:
            logger.info(error_message)
            logger.info("Error Connecting:" + str(errc))
            raise errc
        except requests.exceptions.Timeout as errt:
            logger.info(error_message)
            logger.info("Timeout Error:" + str(errt))
            raise errt
        except requests.exceptions.RequestException as err:
            logger.info(error_message)
            logger.info("An unknown error occured: " + str(err))
            raise err

def keycloak_refresh_token_request(token_endpoint: str, client_id: str, scopes: List[str], refresh_token:str, logger: logging.Logger) -> Dict[str, Any]:
    """Performs the token refresh by making an HTTP post request to the token endpoint
    to obtain new access and refresh tokens.

    Parameters
    ----------
    tokens : Optional[Tokens], optional
        An optional Tokens object containing the refresh token. If not provided,
        the method will use the class variable stored tokens. 

        By default this parameter is None.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing the new access and refresh tokens if the refresh is successful.

    Raises
    ------
    ValueError
        If no refresh token is provided or found in the class token variable. 
    Exception
        If the HTTP request fails a message is displayed with the HTTP status code. Can occur 
        if the refresh token has expired.
    """

    # Perform a refresh grant
    refresh_grant_type = "refresh_token"

    # Required openid connect fields
    data = {
        "grant_type": refresh_grant_type,
        "client_id": client_id,
        "refresh_token": refresh_token,
        "scope": " ".join(scopes)
    }

    logger.info("Attempting to refresh token.")

    # Send API request
    response = requests.post(token_endpoint, data=data)

    if (not response.status_code == 200):
            err_msg = f"The token used for refresh is invalid or has potentially expired. Something went wrong during token refresh. Status code: {response.status_code}."
            logger.error(err_msg)
            raise Exception(err_msg)

    return response.json()


def validate_access_token(public_key: str, access_token: str, logger: logging.Logger) -> bool:
    """Uses the python-jose library to validate current creds.

    In this context, it is basically just checking signature
    and expiry. The tokens are enforced at the API side 
    as well.

    Parameters
    ----------
    tokens : Optional[Tokens], optional
        The tokens object to validate, by default None
    """

    logger.info("Attempting to validate tokens.")

    try:

        jwt_response = jwt.decode(
            access_token,
            public_key,
            algorithms=[ALGORITHMS.RS256],
            options={
                "verify_signature": True,
                "verify_aud": False,
                "exp": True
            }
        )

        token_is_fresh = check_token_expiry_window(jwt_data=jwt_response, logger=logger)

        if not token_is_fresh:
            logger.info(
                "Token is expiring soon and need to be refreshed.")
        else:
            logger.info("Token validation successful.")

        return token_is_fresh

    except JWTError as e:
        logger.info(f"Token Validation Error {e}")
        return False



'''
Created Date: Thursday June 6th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday June 6th 2024 1:39:55 pm +1000
Modified By: Peter Baker
-----
Description: Incomplete Registry API L2 Client.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

28-06-2024 | Parth Kulkarni | Completion of L2 Interface of Registry with General, Admin and Other endpoints. 
18-06-2024 | Peter Baker | Initial structure setup to help dispatch into the various sub types in the L3.
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.models.general import HealthCheckResponse
from provenaclient.utils.config import Config
from enum import Enum
from provenaclient.utils.helpers import *
from provenaclient.clients.client_helpers import *
from provenaclient.utils.registry_endpoints import *
from ProvenaInterfaces.RegistryModels import *
from ProvenaInterfaces.RegistryAPI import *


class GenericRegistryEndpoints(str, Enum):
    GET_HEALTH_CHECK = "/"
    GET_CHECK_ACCESS_CHECK_ADMIN_ACCESS = "/check-access/check-admin-access"
    GET_CHECK_ACCESS_CHECK_GENERAL_ACCESS = "/check-access/check-general-access"
    GET_CHECK_ACCESS_CHECK_READ_ACCESS = "/check-access/check-read-access"
    GET_CHECK_ACCESS_CHECK_WRITE_ACCESS = "/check-access/check-write-access"
    GET_REGISTRY_GENERAL_ABOUT_VERSION = "/registry/general/about/version"
    GET_REGISTRY_GENERAL_FETCH = "/registry/general/fetch"
    POST_REGISTRY_GENERAL_LIST = "/registry/general/list"


class RegistryAdminEndpoints(str, Enum):
    GET_ADMIN_CONFIG = "/admin/config"
    GET_ADMIN_EXPORT = "/admin/export"
    GET_ADMIN_SENTRY_DEBUG = "/admin/sentry-debug"
    POST_ADMIN_IMPORT = "/admin/import"
    POST_ADMIN_RESTORE_FROM_TABLE = "/admin/restore_from_table"

class RegistryAdminClient(ClientService):
    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the RegistryAdminClient with authentication and configuration.

        Parameters
        ----------
        auth: AuthManager
            An abstract interface containing the user's requested auth flow method.
        config: Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_endpoint(self, endpoint: RegistryAdminEndpoints) -> str:
        return f"{self._config.registry_api_endpoint}{endpoint.value}"
    
    def _build_subtype_endpoint(self, action: RegistryAction, item_subtype: ItemSubType) -> str:
        return subtype_action_to_endpoint(
            base=self._config.registry_api_endpoint,
            action=action,
            item_subtype=item_subtype
        )

    async def export_items(self) -> RegistryExportResponse:
        """
        Exports all items from the registry.

        Returns
        -------
        RegistryExportResponse
            The response containing the exported items.
        """
        endpoint = self._build_endpoint(RegistryAdminEndpoints.GET_ADMIN_EXPORT)

        return await parsed_get_request_with_status(
            client=self,
            url=endpoint,
            params=None,
            error_message="Failed to export all items from the registry!",
            model=RegistryExportResponse
        )
    
    async def import_items(self, registry_import_request: RegistryImportRequest) -> RegistryImportResponse:
        """
        Imports items into the registry.

        Parameters
        ----------
        registry_import_request : RegistryImportRequest
            The import request containing the items to import.

        Returns
        -------
        RegistryImportResponse
            The response containing the result of the import operation.
        """
        endpoint = self._build_endpoint(RegistryAdminEndpoints.POST_ADMIN_IMPORT)

        return await parsed_post_request_with_status(
            client=self,
            url=endpoint,
            params=None,
            json_body=py_to_dict(registry_import_request),
            error_message="Failed to import items into the registry!",
            model=RegistryImportResponse
        )
    
    async def restore_items_from_dynamo_table(self, restore_request: RegistryRestoreRequest) -> RegistryImportResponse:
        """
        Restores items from a DynamoDB table into the registry.

        Parameters
        ----------
        restore_request : RegistryRestoreRequest
            The restore request containing the details for restoration.

        Returns
        -------
        RegistryImportResponse
            The response containing the result of the restore operation.
        """
        endpoint = self._build_endpoint(RegistryAdminEndpoints.POST_ADMIN_RESTORE_FROM_TABLE)

        return await parsed_post_request_with_status(
            client=self,
            url=endpoint,
            params=None,
            json_body=py_to_dict(restore_request),
            error_message="Failed to restore items from DynamoDB table into the registry!",
            model=RegistryImportResponse
        )
    
    async def generate_config_file(self, required_only: bool) -> str:
        """
        Generates a nicely formatted .env file of the current required/non-supplied properties.

        Used to quickly bootstrap a local environment or to understand currently deployed API.

        Parameters
        ----------
        required_only : bool
            Whether to include only required properties. By default True.

        Returns
        -------
        str
            The generated .env file content.
        """
        response = await validated_get_request(
            client=self,
            url=self._build_endpoint(RegistryAdminEndpoints.GET_ADMIN_CONFIG),
            error_message="Failed to generate config file",
            params={"required_only": required_only},
        )

        return response.text
    
    async def delete_item(self, id: str, item_subtype: ItemSubType) -> StatusResponse:
        """
        Deletes an item from the registry.

        Parameters
        ----------
        id : str
            The ID of the item to delete.
        item_subtype : ItemSubType
            The subtype of the item to delete.

        Returns
        -------
        StatusResponse
            The status response indicating the result of the deletion.
        """
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.DELETE, item_subtype=item_subtype
        )

        return await parsed_delete_request_with_status(
            client=self,
            params={'id': id},
            error_message=f"Failed to delete item with id {id} and subtype {item_subtype}",
            model=StatusResponse,
            url=endpoint
        )
    
class RegistryGeneralClient(ClientService):

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the RegistryGeneralClient with authentication and configuration.

        Parameters
        ----------
        auth: AuthManager
            An abstract interface containing the user's requested auth flow method.
        config: Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_subtype_endpoint(self, action: RegistryAction, item_subtype: ItemSubType) -> str:
        return subtype_action_to_endpoint(
            base=self._config.registry_api_endpoint,
            action=action,
            item_subtype=item_subtype
        )

    def _build_general_endpoint(self, endpoint: GenericRegistryEndpoints) -> str:
        return f"{self._config.registry_api_endpoint}{endpoint.value}"
    
    async def list_general_registry_items(self, general_list_request: GeneralListRequest) -> PaginatedListResponse:
        """
        Lists general registry items based on filter criteria.

        Parameters
        ----------
        general_list_request : GeneralListRequest
            The request containing filter and sort criteria.

        Returns
        -------
        PaginatedListResponse
            The response containing the paginated list of registry items.
        """
        endpoint = self._build_general_endpoint(endpoint=GenericRegistryEndpoints.POST_REGISTRY_GENERAL_LIST)

        return await parsed_post_request_with_status(
            client=self,
            url=endpoint,
            params=None,
            json_body=py_to_dict(general_list_request),
            error_message=f"General list fetch failed!",
            model=PaginatedListResponse
        )
    
    async def general_fetch_item(self, id: str) -> UntypedFetchResponse:
        """
        Fetches a general item from the registry.

        Parameters
        ----------
        id : str
            The ID of the item to fetch.

        Returns
        -------
        UntypedFetchResponse
            The fetch response containing the item details.
        """
        endpoint = self._build_general_endpoint(endpoint=GenericRegistryEndpoints.GET_REGISTRY_GENERAL_FETCH)

        return await parsed_get_request_with_status(
            client=self,
            url=endpoint,
            params={"id": id},
            error_message=f"Failed to fetch item with id {id} from general registry!",
            model=UntypedFetchResponse
        )
    
    async def get_current_provena_version(self) -> VersionResponse:
        """
        Gets the current Provena version.

        Returns
        -------
        VersionResponse
            The response containing the current Provena version.
        """
        endpoint = self._build_general_endpoint(endpoint=GenericRegistryEndpoints.GET_REGISTRY_GENERAL_ABOUT_VERSION)

        return await parsed_get_request(
            client=self,
            url=endpoint,
            params=None,
            error_message="Failed to fetch the current Provena version of your instance.",
            model=VersionResponse
        )

class RegistryClient(ClientService):
    # Sub clients
    admin: RegistryAdminClient
    general: RegistryGeneralClient

    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the RegistryClient with authentication and configuration.

        Parameters
        ----------
        auth: AuthManager
            An abstract interface containing the user's requested auth flow method.
        config: Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

        # Sub clients
        self.admin = RegistryAdminClient(auth=auth, config=config)
        self.general = RegistryGeneralClient(auth=auth, config=config)

    # Function to get the endpoint URL
    def _build_subtype_endpoint(self, action: RegistryAction, item_subtype: ItemSubType) -> str:
        return subtype_action_to_endpoint(
            base=self._config.registry_api_endpoint,
            action=action,
            item_subtype=item_subtype
        )

    def _build_general_endpoint(self, endpoint: GenericRegistryEndpoints) -> str:
        return f"{self._config.registry_api_endpoint}{endpoint.value}"

    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns
        -------
        HealthCheckResponse
            Response
        """
        return await parsed_get_request(
            client=self,
            url=self._build_general_endpoint(GenericRegistryEndpoints.GET_HEALTH_CHECK),
            error_message="Health check failed!",
            params={},
            model=HealthCheckResponse
        )

    async def fetch_item(self, id: str, item_subtype: ItemSubType, fetch_response_model: Type[BaseModelType], seed_allowed: Optional[bool] = None) -> BaseModelType:
        """
        Ascertains the correct endpoint based on the subtype provided, then runs the fetch operation, parsing the data as the specified model.

        Parameters
        ----------
        id : str
            The id of the item to fetch.
        item_subtype : ItemSubType
            The subtype of the item to fetch.
        fetch_response_model : Type[BaseModelType]
            The response pydantic model to parse as e.g. OrganisationFetchResponse
        seed_allowed : Optional[bool], optional
            Should the endpoint throw an error if the item is a seed item?  Defaults to None.

        Returns
        -------
        BaseModelType
            The fetch response parsed as the specified model.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.FETCH, item_subtype=item_subtype)

        # fetch the item from the subtype specific endpoint
        return await parsed_get_request_with_status(
            client=self,
            params={'id': id, 'seed_allowed': seed_allowed},
            error_message=f"Failed to fetch item with id {id} and subtype {item_subtype}.",
            model=fetch_response_model,
            url=endpoint,
        )

    async def update_item(self, id: str, reason: Optional[str], item_subtype: ItemSubType, domain_info: DomainInfoBase, update_response_model: Type[BaseModelType]) -> BaseModelType:
        """
        Ascertains the correct endpoint then runs the update operation on an existing item by providing new domain info.

        Parameters
        ----------
        id : str
            The id of item to update.
        reason : Optional[str]
            The reason for updating, if any
        item_subtype : ItemSubType
            The subtype to update
        domain_info : DomainInfoBase
            The domain info to replace existing item with
        update_response_model : Type[BaseModelType]
            The response model to parse e.g. StatusResponse

        Returns
        -------
        BaseModelType
            The response model parsed
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.UPDATE, item_subtype=item_subtype)

        # fetch the item from the subtype specific endpoint
        return await parsed_put_request_with_status(
            client=self,
            params={'id': id, 'reason': reason},
            json_body=py_to_dict(domain_info),
            error_message=f"Failed to update item with id {id} and subtype {item_subtype}.",
            model=update_response_model,
            url=endpoint,
        )

    async def list_items(self, list_items_payload: GeneralListRequest, item_subtype: ItemSubType, update_model_response: Type[BaseModelType]) -> BaseModelType:
        """
        Lists items within the registry based on filter criteria.

        Parameters
        ----------
        list_items_payload : GeneralListRequest
            The request containing filter and sort criteria.
        item_subtype : ItemSubType
            The subtype of the items to list.
        update_model_response : Type[BaseModelType]
            The response model to parse.

        Returns
        -------
        BaseModelType
            The response containing the list of items.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.LIST, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_post_request_with_status(
            client=self,
            params=None,
            json_body=py_to_dict(list_items_payload),
            error_message=f"Failed to list items for {item_subtype}",
            model=update_model_response,
            url=endpoint
        )

    async def seed_item(self, item_subtype: ItemSubType, seed_model_response: Type[BaseModelType]) -> BaseModelType:
        """
        Seeds an item in the registry.

        Parameters
        ----------
        item_subtype : ItemSubType
            The subtype of the item to seed.
        seed_model_response : Type[BaseModelType]
            The response model to parse.

        Returns
        -------
        BaseModelType
            The response containing the details of the seeded item.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.SEED, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_post_request_with_status(
            client=self,
            params=None,
            json_body=None,
            error_message=f"Failed to seed items for {item_subtype}",
            model=seed_model_response,
            url=endpoint
        )

    async def revert_item(self, revert_request: ItemRevertRequest, item_subtype: ItemSubType) -> ItemRevertResponse:
        """
        Reverts an item in the registry.

        Parameters
        ----------
        revert_request : ItemRevertRequest
            The revert request.
        item_subtype : ItemSubType
            The subtype of the item to revert.

        Returns
        -------
        ItemRevertResponse
            The revert response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.REVERT, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_put_request_with_status(
            client=self,
            params=None,
            json_body=py_to_dict(revert_request),
            error_message=f"Failed to revert items for {item_subtype}",
            model=ItemRevertResponse,
            url=endpoint
        )

    async def create_item(self, create_item_request: DomainInfoBase, item_subtype: ItemSubType, create_response_model: Type[BaseModelType]) -> BaseModelType:
        """
        Creates an item in the registry.

        Parameters
        ----------
        create_item_request : DomainInfoBase
            The domain information required to create the item.
        item_subtype : ItemSubType
            The subtype of the item to create.
        create_response_model : Type[BaseModelType]
            The response model to parse.

        Returns
        -------
        BaseModelType
            The response containing the details of the created item.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.CREATE, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_post_request_with_status(
            client=self,
            params=None,
            json_body=py_to_dict(create_item_request),
            error_message=f"Failed to create items for {item_subtype}",
            model=create_response_model,
            url=endpoint
        )

    async def get_schema(self, item_subtype: ItemSubType) -> JsonSchemaResponse:
        """
        Gets the schema for the item subtype.

        Parameters
        ----------
        item_subtype : ItemSubType
            The subtype of the item to get the schema for.

        Returns
        -------
        JsonSchemaResponse
            The JSON schema response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.SCHEMA, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_get_request_with_status(
            client=self,
            params=None,
            error_message=f"Failed to get schema for {item_subtype}",
            model=JsonSchemaResponse,
            url=endpoint
        )

    async def validate_item(self, validate_request: DomainInfoBase, item_subtype: ItemSubType) -> StatusResponse:
        """
        Validates an item in the registry.

        Parameters
        ----------
        validate_request : DomainInfoBase
            The domain information of the item to be validated.
        item_subtype : ItemSubType
            The subtype of the item to validate.

        Returns
        -------
        StatusResponse
            The status response indicating the result of the validation.
        """

        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.VALIDATE, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_post_request_with_status(
            client=self,
            params=None,
            json_body=py_to_dict(validate_request),
            error_message=f"Failed to validate item for {item_subtype}",
            model=StatusResponse,
            url=endpoint
        )

    async def evaluate_auth_access(self, id: str, item_subtype: ItemSubType) -> DescribeAccessResponse:
        """
        Evaluates the auth access for an item.

        Parameters
        ----------
        id : str
            The ID of the item to evaluate auth access for.
        item_subtype : ItemSubType
            The subtype of the item to evaluate auth access for.

        Returns
        -------
        DescribeAccessResponse
            The describe access response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.AUTH_EVALUATE, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_get_request(
            client=self,
            params={"id": id},
            error_message=f"Failed to evaluate auth access for {item_subtype}",
            model=DescribeAccessResponse,
            url=endpoint
        )

    async def get_auth_configuration(self, id: str, item_subtype: ItemSubType) -> AccessSettings:
        """
        Gets the auth configuration for an item.

        Parameters
        ----------
        id : str
            The ID of the item to get auth configuration for.
        item_subtype : ItemSubType
            The subtype of the item to get auth configuration for.

        Returns
        -------
        AccessSettings
            The access settings.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.AUTH_CONFIGURATION, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_get_request(
            client=self,
            params={"id": id},
            error_message=f"Failed to get auth config for {item_subtype}",
            model=AccessSettings,
            url=endpoint
        )

    async def modify_auth_configuration(self, id: str, auth_change_request: AccessSettings, item_subtype: ItemSubType) -> StatusResponse:
        """
        Modifies the auth configuration for an item.

        Parameters
        ----------
        id : str
            The ID of the item to modify auth configuration for.
        auth_change_request : AccessSettings
            The auth change request.
        item_subtype : ItemSubType
            The subtype of the item to modify auth configuration for.

        Returns
        -------
        StatusResponse
            The status response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.AUTH_CONFIGURATION, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_put_request_with_status(
            client=self,
            params={"id": id},
            json_body=py_to_dict(auth_change_request),
            error_message=f"Failed to modify auth config for {item_subtype}",
            model=StatusResponse,
            url=endpoint
        )

    async def get_auth_roles(self, item_subtype: ItemSubType) -> AuthRolesResponse:
        """
        Gets the auth roles for the item subtype.

        Parameters
        ----------
        item_subtype : ItemSubType
            The subtype of the item to get auth roles for.

        Returns
        -------
        AuthRolesResponse
            The auth roles response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.AUTH_ROLES, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_get_request(
            client=self,
            params=None,
            error_message=f"Failed to get auth roles for {item_subtype}",
            model=AuthRolesResponse,
            url=endpoint
        )

    async def lock_resource(self, lock_resource_request: LockChangeRequest, item_subtype: ItemSubType) -> StatusResponse:
        """
        Locks a resource in the registry.

        Parameters
        ----------
        lock_resource_request : LockChangeRequest
            The lock resource request.
        item_subtype : ItemSubType
            The subtype of the resource to lock.

        Returns
        -------
        StatusResponse
            The status response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.LOCK, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_put_request_with_status(
            client=self,
            params=None,
            json_body=py_to_dict(lock_resource_request),
            error_message=f"Failed to lock resource for {item_subtype}",
            model=StatusResponse,
            url=endpoint
        )

    async def unlock_resource(self, unlock_resource_request: LockChangeRequest, item_subtype: ItemSubType) -> StatusResponse:
        """
        Unlocks a resource in the registry.

        Parameters
        ----------
        unlock_resource_request : LockChangeRequest
            The unlock resource request.
        item_subtype : ItemSubType
            The subtype of the resource to unlock.

        Returns
        -------
        StatusResponse
            The status response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.UNLOCK, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_put_request_with_status(
            client=self,
            params=None,
            json_body=py_to_dict(unlock_resource_request),
            error_message=f"Failed to unlock resource for {item_subtype}",
            model=StatusResponse,
            url=endpoint
        )

    async def get_lock_history(self, handle_id: str, item_subtype: ItemSubType) -> LockHistoryResponse:
        """
        Gets the lock history for an item.

        Parameters
        ----------
        handle_id : str
            The handle ID of the item to get lock history for.
        item_subtype : ItemSubType
            The subtype of the item to get lock history for.

        Returns
        -------
        LockHistoryResponse
            The lock history response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.LOCK_HISTORY, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_get_request_with_status(
            client=self,
            params={"id": handle_id},
            error_message=f"Failed to get lock history for id {handle_id} for subtype {item_subtype}",
            model=LockHistoryResponse,
            url=endpoint
        )

    async def get_lock_status(self, id: str, item_subtype: ItemSubType) -> LockStatusResponse:
        """
        Gets the lock status for an item.

        Parameters
        ----------
        id : str
            The item ID.
        item_subtype : ItemSubType
            The subtype of the item to get lock status for.

        Returns
        -------
        LockStatusResponse
            The lock status response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.LOCK_HISTORY, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_get_request_with_status(
            client=self,
            params={"id": id},
            error_message=f"Failed to get lock status for id {id} with subtype {item_subtype}",
            model=LockStatusResponse,
            url=endpoint
        )

    async def version(self, version_request: VersionRequest, item_subtype: ItemSubType) -> VersionResponse:
        """
        Versions an item in the registry.

        Parameters
        ----------
        version_request : VersionRequest
            The version request containing the version details.
        item_subtype : ItemSubType
            The subtype of the item to version.

        Returns
        -------
        VersionResponse
            The version response.
        """
        # determine endpoint
        endpoint = self._build_subtype_endpoint(
            action=RegistryAction.VERSION, item_subtype=item_subtype
        )

        # fetch item from the subtype specific endpoint
        return await parsed_post_request(
            client=self,
            params=None,
            json_body=py_to_dict(version_request),
            error_message=f"Failed to complete versioning for subtype {item_subtype}",
            model=VersionResponse,
            url=endpoint
        )


'''
Created Date: Thursday May 30th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday May 30th 2024 10:17:36 am +1000
Modified By: Peter Baker
-----
Description: Defines L3 module level helpers, notably a common interface for the ModuleService.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Noting this might not be a great place to host this interface.
'''

from abc import ABC
from provenaclient.auth import AuthManager
from provenaclient.utils.config import Config


class ModuleService(ABC):
    """
    This class interface just captures that the client has an instantiated auth
    manager which allows for helper functions abstracted for L3 clients.
    """
    _auth: AuthManager
    _config: Config



'''
Created Date: Monday June 17th 2024 +1000
Author: Peter Baker
-----
Last Modified: Monday June 17th 2024 4:45:39 pm +1000
Modified By: Peter Baker
-----
Description: Provenance API L3 module. Includes the ProvAPI sub module. Contains IO helper functions for writing/reading files.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import ProvClient
from provenaclient.utils.exceptions import *
from provenaclient.modules.module_helpers import *
from provenaclient.utils.helpers import read_file_helper, write_file_helper, get_and_validate_file_path
from typing import List
from provenaclient.models.general import HealthCheckResponse
from ProvenaInterfaces.ProvenanceAPI import LineageResponse, ModelRunRecord, ConvertModelRunsResponse, RegisterModelRunResponse, RegisterBatchModelRunRequest, RegisterBatchModelRunResponse
from ProvenaInterfaces.RegistryAPI import ItemModelRun
from ProvenaInterfaces.SharedTypes import StatusResponse

# L3 interface.

PROV_API_DEFAULT_SEARCH_DEPTH = 3
DEFAULT_CONFIG_FILE_NAME = "prov-api.env"

class ProvAPIAdminSubModule(ModuleService):
    _prov_api_client: ProvClient

    def __init__(self, auth: AuthManager, config: Config, prov_api_client: ProvClient) -> None:
        """
        Admin sub module of the Prov API providing functionality
        for the admin endpoints.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        auth_client: AuthClient
            The instantiated auth client
        """
        self._auth = auth
        self._config = config

        # Clients related to the prov_api scoped as private.
        self._prov_api_client = prov_api_client

    async def generate_config_file(self, required_only: bool = True, file_path: Optional[str] = None, write_to_file: bool = False) -> str:
        """Generates a nicely formatted .env file of the current required/non supplied properties 
        Used to quickly bootstrap a local environment or to understand currently deployed API.

        Parameters
        ----------
        required_only : bool, optional
            By default True
        file_path: str, optional
            The path you want to save the config file at WITH the file name. If you don't specify a path
            this will be saved in a relative directory.
        write_to_file: bool, By default False
            A boolean flag to indicate whether you want to save the config response to a file
            or not.

        Returns
        ----------
        str: Response containing the config text.

        """

        file_path = get_and_validate_file_path(file_path=file_path, write_to_file=write_to_file, default_file_name=DEFAULT_CONFIG_FILE_NAME)

        config_text: str = await self._prov_api_client.admin.generate_config_file(required_only=required_only)

        if config_text is None:
            raise ValueError(f"No data returned for generate config file endpoint.")

        # Write to file if config text is not None, write to file is True and file path is not None.
        if write_to_file:
            if file_path is None:
                raise ValueError("File path is not set for writing the CSV.")
            write_file_helper(file_path=file_path, content=config_text)

        return config_text
    
    async def store_record(self, registry_record: ItemModelRun, validate_record: bool = True) -> StatusResponse:
        """An admin only endpoint which enables the reupload/storage of an existing completed provenance record.

        Parameters
        ----------
        registry_record : ItemModelRun
            The completed registry record for the model run.
        validate_record: bool
            Optional Should the ids in the payload be validated?, by default True

        Returns
        -------
        StatusResponse
            A status response indicating the success of the request and any other details.
        """

        return await self._prov_api_client.admin.store_record(registry_record=registry_record, validate_record = validate_record)
        
    async def store_multiple_records(self, registry_record: List[ItemModelRun], validate_record: bool = True) -> StatusResponse:
        """An admin only endpoint which enables the reupload/storage of an existing but multiple completed provenance record.

        Parameters
        ----------
        registry_record : List[ItemModelRun]
            List of the completed registry record for the model run validate_record
        validate_record: bool
            Optional Should the ids in the payload be validated?, by default True

        Returns
        -------
        StatusResponse
            A status response indicating the success of the request and any other details.
        """

        return await self._prov_api_client.admin.store_multiple_records(registry_record=registry_record, validate_record = validate_record)

    async def store_all_registry_records(self, validate_record: bool = True) -> StatusResponse:
        """Applies the store record endpoint action across a list of ItemModelRuns '
           which is found by querying the registry model run list endpoint directly.

        Parameters
        ----------
        validate_record : bool
            Optional Should the ids in the payload be validated?, by default True


        Returns
        -------
        StatusResponse
            A status response indicating the success of the request and any other details.
        """

        return await self._prov_api_client.admin.store_all_registry_records(validate_record=validate_record)



class Prov(ModuleService):
    _prov_client: ProvClient

    def __init__(self, auth: AuthManager, config: Config, prov_client: ProvClient) -> None:
        """Initialises a new datastore object, which sits between the user and the datastore api operations.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance. 
        datastore_client : DatastoreClient
            This client interacts with the Datastore API's.
        """
        self._auth = auth
        self._config = config

        # Clients related to the prov-api scoped as private.
        self._prov_api_client = prov_client

        # Submodules 
        self.admin = ProvAPIAdminSubModule(auth, config, prov_client)
    
    async def get_health_check(self) -> HealthCheckResponse:
        """Checks the health status of the PROV-API.

        Returns
        -------
        HealthCheckResponse
            Response containing the PROV-API health information.
        """

        return await self._prov_api_client.get_health_check()

    async def explore_upstream(self, starting_id: str, depth: int = PROV_API_DEFAULT_SEARCH_DEPTH) -> LineageResponse:
        """Explores in the upstream direction (inputs/associations) 
        starting at the specified node handle ID. 
        The search depth is bounded by the depth parameter which has a default maximum of 100.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the upstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await self._prov_api_client.explore_upstream(starting_id=starting_id, depth=depth) 
    
    async def explore_downstream(self, starting_id: str, depth: int = PROV_API_DEFAULT_SEARCH_DEPTH) -> LineageResponse:
        """Explores in the downstream direction (inputs/associations) 
        starting at the specified node handle ID. 
        The search depth is bounded by the depth parameter which has a default maximum of 100.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the downstream direction, by default 100

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await self._prov_api_client.explore_downstream(starting_id=starting_id, depth=depth)
    
    async def get_contributing_datasets(self, starting_id: str, depth:int = PROV_API_DEFAULT_SEARCH_DEPTH) -> LineageResponse:
        """Fetches datasets (inputs) which involved in a model run
        naturally in the upstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the upstream direction, by default 100

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await self._prov_api_client.get_contributing_datasets(starting_id=starting_id, depth=depth)
    
    async def get_effected_datasets(self, starting_id: str, depth:int = PROV_API_DEFAULT_SEARCH_DEPTH) -> LineageResponse:
        """Fetches datasets (outputs) which are derived from the model run
        naturally in the downstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the downstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await self._prov_api_client.get_effected_datasets(starting_id=starting_id, depth=depth)

    async def get_contributing_agents(self, starting_id: str, depth:int = PROV_API_DEFAULT_SEARCH_DEPTH) -> LineageResponse:
        """Fetches agents (organisations or peoples) that are involved or impacted by the model run.
        naturally in the upstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the upstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await self._prov_api_client.get_contributing_agents(starting_id=starting_id, depth=depth)

    async def get_effected_agents(self, starting_id: str, depth:int = PROV_API_DEFAULT_SEARCH_DEPTH) -> LineageResponse:
        """Fetches agents (organisations or peoples) that are involved or impacted by the model run.
        naturally in the downstream direction.

        Parameters
        ----------
        starting_id : str
            The ID of the entity to start at.
        depth : int, optional
            The depth to traverse in the downstream direction, by default 100.

        Returns
        -------
        LineageResponse
            A response containing the status, node count, and networkx serialised graph response.
        """

        return await self._prov_api_client.get_effected_agents(starting_id=starting_id, depth=depth)

    async def register_batch_model_runs(self, batch_model_run_payload: RegisterBatchModelRunRequest) -> RegisterBatchModelRunResponse:
        """This function allows you to register multiple model runs in one go (batch) asynchronously.

        Note: You can utilise the returned session ID to poll on 
        the JOB API to check status of the model run registration(s).

        Parameters
        ----------
        batch_model_run_payload : RegisterBatchModelRunRequest
            A list of model runs (ModelRunRecord objects)

        Returns
        -------
        RegisterBatchModelRunResponse
            The job session id derived from job-api for the model-run batch. 
        """

        return await self._prov_api_client.register_batch_model_runs(model_run_batch_payload = batch_model_run_payload)

    async def register_model_run(self, model_run_payload: ModelRunRecord) -> RegisterModelRunResponse:
        """Asynchronously registers a single model run.

        Note: You can utilise the returned session ID to poll on 
        the JOB API to check status of the model run registration.

        Parameters
        ----------
        model_run_payload : ModelRunRecord
            Contains information needed for the 
            model run such as workflow template,
            inputs, outputs, description etc. 

        Returns
        -------
        RegisterModelRunResponse
            The job session id derived from job-api for the model-run. 
        """

        return await self._prov_api_client.register_model_run(model_run_payload=model_run_payload)
    
    async def generate_csv_template(self, workflow_template_id: str, file_path: Optional[str] = None, write_to_csv: bool = False) -> str:
        """Generates a model run csv template to be utilised 
        for creating model runs through csv format..

        Parameters
        ----------
        workflow_template_id : str
            An ID of a created and existing model run workflow template.
        path_to_save_csv: str, optional 
            The path you want to save the csv file at WITH csv file name. If you don't specify a path
            this will be saved in a relative directory.
        write_to_csv: bool, By default False
            A boolean flag to indicate whether you want to save the template to a csv file
            or not.

        Returns
        ----------
        str: Response containing the csv template text (encoded in a csv format). 
        
        """

        file_path = get_and_validate_file_path(file_path=file_path, write_to_file=write_to_csv, default_file_name=workflow_template_id + ".csv")

        csv_text = await self._prov_api_client.generate_csv_template(workflow_template_id=workflow_template_id)

        if csv_text is None:
            raise ValueError(f"No data returned for generate CSV template workflow template ID {workflow_template_id}")

        # Write to file if CSV content is returned and write_to_csv is True and file path is assigned.
        if write_to_csv:
            if file_path is None:
                raise ValueError("File path is not set for writing the CSV.")
            write_file_helper(file_path=file_path, content=csv_text)

        return csv_text
    

    async def convert_model_runs(self,model_run_content: str) -> ConvertModelRunsResponse: 
        """Converts model run with model_run_content provided as a string.
        
        Parameters
        ----------
        model_run_content : str
            The model run information containing
            the necessary parameters for model run lodge.

        Returns
        -------
        ConvertModelRunsResponse
            Returns the model run information in an interactive python
            datatype.

        Raises
        ------
        Exception
            Exception raised when converting string to bytes.
        """

        response = await self._prov_api_client.convert_model_runs_to_csv(csv_file_contents=model_run_content)

        return response

    async def convert_model_runs_to_csv_with_file(self, file_path: str) -> ConvertModelRunsResponse:
        """Reads a CSV file, and it's defined model run contents
        and lodges a model run.

        Parameters
        ----------
        file_path : str
            The path of an existing created CSV file containing
            the necessary parameters for model run lodge.

        Returns
        -------
        ConvertModelRunsResponse
            Returns the model run information in an interactive python
            datatype.

        """

        file_content = read_file_helper(file_path = file_path)

        response = await self._prov_api_client.convert_model_runs_to_csv(csv_file_contents=file_content)
        return response

    async def regenerate_csv_from_model_run_batch(self, batch_id: str, file_path: Optional[str] = None, write_to_csv: bool = False) -> str:
        """Regenerate/create a csv file containing model 
        run information from a model run batch job.

        The batch id must exist in the system.

        Parameters
        ----------
        batch_id : str
            Obtained from creating a batch model run.
        file_path: str, optional 
            The path you want to save the csv file at WITH CSV file name. If you don't specify a path
            this will be saved in a relative directory.
        write_to_csv: bool, By default False
            A boolean flag to indicate whether you want to save the template to a csv file
            or not.

        Returns
        ----------
        str: Response containing the model run information (encoded in csv format).
        
        """

        file_path = get_and_validate_file_path(file_path=file_path, write_to_file=write_to_csv, default_file_name=batch_id + ".csv")

        csv_text: str = await self._prov_api_client.regenerate_csv_from_model_run_batch(batch_id=batch_id)

        if csv_text is None:
            raise ValueError(f"No data returned for batch ID {batch_id}")
        
        # Write to file if CSV content is returned and write_to_csv is True and file path is assigned.
        if write_to_csv:
            if file_path is None:
                raise ValueError("File path is not set for writing the CSV.")
            write_file_helper(file_path=file_path, content=csv_text)

        return csv_text


'''
Created Date: Thursday May 30th 2024 +1000
Author: Peter Baker
-----
Last Modified: Thursday May 30th 2024 10:17:36 am +1000
Modified By: Peter Baker
-----
Description: Search API L3 Module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import SearchClient
from provenaclient.clients.client_helpers import *
from ProvenaInterfaces.RegistryModels import ItemSubType
from ProvenaInterfaces.SearchAPI import QueryResults
from typing import Optional

# L3 interface.

DEFAULT_SEARCH_LIMIT = 25


class Search(ClientService):
    _search_client: SearchClient

    def __init__(self, auth: AuthManager, config: Config, search_client: SearchClient) -> None:
        """

        Initialises a new search object, which sits between the user and the
        search api operations.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        search_client : SearchClient
            This client interacts with the Search Client's API's.
        """
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._search_client = search_client

    async def search_registry(self, query: str, limit: Optional[int], subtype_filter: Optional[ItemSubType]) -> QueryResults:
        """

        Searches the registry for given query, limit and subtype.

        Args:
            query (str): The query to make
            limit (Optional[int]): The query limit
            subtype_filter (Optional[ItemSubType]): The subtype to filter by if any

        Returns:
            QueryResults: Results - ids and scores
        """
        return await self._search_client.search_registry(
            limit=limit, query=query, subtype_filter=subtype_filter
        )



from provenaclient.clients.datastore_client import DatastoreClient
from provenaclient.clients.search_client import SearchClient
from provenaclient.clients.auth_client import AuthClient
from provenaclient.clients.prov_client import ProvClient
from provenaclient.clients.registry_client import RegistryClient
from provenaclient.clients.job_client import JobAPIClient
from provenaclient.clients.id_service_client import IdServiceClient


'''
Created Date: Friday May 31st 2024 +1000
Author: Peter Baker
-----
Last Modified: Friday May 31st 2024 9:50:26 am +1000
Modified By: Peter Baker
-----
Description: A L3 module which is setup to interact more nicely with the Username Person Link service.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Initial ideas included but is not finished/ready. Blocked by Registry API functionality.
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.modules.module_helpers import ModuleService
from provenaclient.clients import AuthClient

# L3 interface.

class Link(ModuleService):
    _auth_client: AuthClient

    def __init__(self, auth: AuthManager, config: Config, auth_client: AuthClient) -> None:
        """
        The link service L3 is a set of helper functions to interact more intuitively with the link service. 

        For example, combining looking up user link details with resolving the person. 

        Or could help to create and link person in one action.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance.
        auth_client: AuthClient 
            L2 auth client used to help with link service
        """
        # Module service
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._auth_client = auth_client

    async def get_linked_person(self) -> None:
        # fetches current user link then loads person
        # TODO pseudocode below to complete once registry is implemented
        # get link
        link_result = await self._auth_client.get_link_lookup_username()

        # then load person with resulting id
        person_id = link_result.person_id
        if person_id is None:
            return None
        else:
            # TODO
            # return await self._registry_client.person.fetch(id=person_id)
            return None

    async def create_and_link_person(self) -> None:
        # TODO
        # check user is not already linked
        # check there is not already a person in registry which matches first/last name or which has matching username
        # then create person with specified details
        # then link person
        return None



'''
Created Date: Friday May 31st 2024 +1000
Author: Peter Baker
-----
Last Modified: Friday May 31st 2024 9:50:26 am +1000
Modified By: Peter Baker
-----
Description: ID/Handle service L2 Client
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------

18-06-2024 | Peter Baker | Noting that the models published for the endpoints are not being detected in Swagger docs due to a type Alias being used
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from enum import Enum
from provenaclient.utils.helpers import *
from provenaclient.clients.client_helpers import *
from ProvenaInterfaces.HandleAPI import *


class IdServiceEndpoints(str, Enum):
    """An ENUM containing the handle api endpoints."""
    GET_ADMIN_CONFIG = "/admin/config"
    GET_ADMIN_SENTRY_DEBUG = "/admin/sentry-debug"
    GET_CHECK_ACCESS_CHECK_GENERAL_ACCESS = "/check-access/check-general-access"
    GET_CHECK_ACCESS_CHECK_ADMIN_ACCESS = "/check-access/check-admin-access"
    GET_CHECK_ACCESS_CHECK_READ_ACCESS = "/check-access/check-read-access"
    GET_CHECK_ACCESS_CHECK_WRITE_ACCESS = "/check-access/check-write-access"
    POST_HANDLE_MINT = "/handle/mint"
    POST_HANDLE_ADD_VALUE = "/handle/add_value"
    POST_HANDLE_ADD_VALUE_BY_INDEX = "/handle/add_value_by_index"
    GET_HANDLE_GET = "/handle/get"
    GET_HANDLE_LIST = "/handle/list"
    PUT_HANDLE_MODIFY_BY_INDEX = "/handle/modify_by_index"
    POST_HANDLE_REMOVE_BY_INDEX = "/handle/remove_by_index"
    GET_HEALTH_CHECK = "/"

# L2 interface.


class IdServiceClient(ClientService):
    def __init__(self, auth: AuthManager, config: Config) -> None:
        """Initialises the IdServiceClient with authentication and configuration.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow method.
        config : Config
            A config object which contains information related to the Provena instance.
        """
        self._auth = auth
        self._config = config

    def _build_endpoint(self, endpoint: IdServiceEndpoints) -> str:
        return self._config.handle_service_api_endpoint + endpoint.value

    async def mint(self, body: MintRequest) -> Handle:
        """
        
        Mints a new handle with given properties.

        Args:
            body (MintRequest): The payload

        Returns:
            Handle: The resulting handle object
        """
        return await parsed_post_request(
            client=self,
            params={},
            json_body=py_to_dict(body),
            url=self._build_endpoint(IdServiceEndpoints.POST_HANDLE_MINT),
            model=Handle,
            error_message=f"Failed to mint new handle."
        )
        
    async def add_value(self, body: AddValueRequest) -> Handle:
        """
        Adds value to a handle.

        Args:
            body (AddValueRequest): The value to add

        Returns:
            Handle: The handle object
        """
        return await parsed_post_request(
            client=self,
            params={},
            json_body=py_to_dict(body),
            url=self._build_endpoint(IdServiceEndpoints.POST_HANDLE_ADD_VALUE),
            model=Handle,
            error_message=f"Failed to add value to handle."
        )
        
    async def add_value_by_index(self, body: AddValueIndexRequest) -> Handle:
        """
        Adds a value at specified index

        Args:
            body (AddValueIndexRequest): The value to add

        Returns:
            Handle: The handle object
        """
        return await parsed_post_request(
            client=self,
            params={},
            json_body=py_to_dict(body),
            url=self._build_endpoint(IdServiceEndpoints.POST_HANDLE_ADD_VALUE_BY_INDEX),
            model=Handle,
            error_message=f"Failed to add value by index to handle."
        )

    async def list(self) -> ListResponse:
        """
        Lists all handles under domain.

        Returns:
            ListResponse: The response list
        """
        return await parsed_get_request(
            client=self,
            params={},
            url=self._build_endpoint(IdServiceEndpoints.GET_HANDLE_LIST),
            model=ListResponse,
            error_message=f"Failed to list handles."
        )
        
    async def modify_by_index(self, body: ModifyRequest) -> Handle:
        """
        Modifies existing handle value at specified index

        Args:
            body (ModifyRequest): The request to modify

        Returns:
            Handle: The handle object
        """ 
        return await parsed_post_request(
            client=self,
            params={},
            json_body=py_to_dict(body),
            url=self._build_endpoint(IdServiceEndpoints.PUT_HANDLE_MODIFY_BY_INDEX),
            model=Handle,
            error_message=f"Failed to modify property by index."
        )
        
    async def remove_by_index(self, body: RemoveRequest) -> Handle:
        """
        
        Removes value at specified index

        Args:
            body (RemoveRequest): The removal request

        Returns:
            Handle: The handle object
        """
        return await parsed_post_request(
            client=self,
            params={},
            json_body=py_to_dict(body),
            url=self._build_endpoint(IdServiceEndpoints.POST_HANDLE_REMOVE_BY_INDEX),
            model=Handle,
            error_message=f"Failed to remove property by index."
        )


'''
Created Date: Friday May 31st 2024 +1000
Author: Peter Baker
-----
Last Modified: Friday May 31st 2024 9:50:26 am +1000
Modified By: Peter Baker
-----
Description: Auth API L3 Module.
-----
HISTORY:
Date      	By	Comments
----------	---	---------------------------------------------------------
'''

from provenaclient.auth.manager import AuthManager
from provenaclient.utils.config import Config
from provenaclient.clients import AuthClient
from provenaclient.modules.module_helpers import *
from provenaclient.models import HealthCheckResponse
from ProvenaInterfaces.AuthAPI import *

# L3 interface.


class AdminAuthSubModule(ModuleService):
    _auth_client: AuthClient

    def __init__(self, auth: AuthManager, config: Config, auth_client: AuthClient) -> None:
        """
        Admin sub module of the Auth API functionality

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        auth_client: AuthClient
            The instantiated auth client
        """
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._auth_client = auth_client

    async def get_all_pending_request_history(self) -> AccessRequestList:
        """

        Gets all requests with pending status

        Returns:
            AccessRequestList: The response object
        """
        return await self._auth_client.admin.get_all_pending_request_history()

    async def get_all_request_history(self) -> AccessRequestList:
        """

        Gets all requests

        Returns:
            AccessRequestList: The response object
        """
        return await self._auth_client.admin.get_all_request_history()

    async def get_user_pending_request_history(self, username: str) -> AccessRequestList:
        """

        Gets pending requests for specified user

        Args:
            username (str): Username to query

        Returns:
            AccessRequestList: The response list
        """
        return await self._auth_client.admin.get_user_pending_request_history(username=username)

    async def get_user_request_history(self, username: str) -> AccessRequestList:
        """

        Gets all requests for specified user

        Args:
            username (str): Username to query

        Returns:
            AccessRequestList: The response list
        """
        return await self._auth_client.admin.get_user_request_history(username=username)

    async def post_add_note(self, note: RequestAddNote) -> None:
        """
        Adds a note to an existing request

        Args:
            note (RequestAddNote): Payload incl note info
        """
        return await self._auth_client.admin.post_add_note(
            note=note
        )

    async def post_change_request_state(self, send_email_alert: bool, change: AccessRequestStatusChange) -> ChangeStateStatus:
        """

        Change state of a request. 

        Args:
            send_email_alert (bool): Should trigger email alert?
            change (AccessRequestStatusChange): The details of change

        Returns:
            ChangeStateStatus: The response object
        """
        return await self._auth_client.admin.post_change_request_state(
            change=change,
            send_email_alert=send_email_alert
        )

    async def get_list_groups(self) -> ListGroupsResponse:
        """
        
        Gets a list of groups

        Returns:
            ListGroupsResponse: List of groups
        """
        return await self._auth_client.admin.get_list_groups()

    async def get_describe_group(self, group_id: str) -> DescribeGroupResponse:
        """
        
        Describes a group by ID

        Args:
            group_id (str): The group

        Returns:
            DescribeGroupResponse: Description
        """
        return await self._auth_client.admin.get_describe_group(group_id=group_id)

    async def get_list_members(self, group_id: str) -> ListMembersResponse:
        """
        Lists members of group

        Args:
            group_id (str): The gruop

        Returns:
            ListMembersResponse: The list of members
        """
        return await self._auth_client.admin.get_list_members(group_id=group_id)

    async def get_list_group_membership(self, username: str) -> ListUserMembershipResponse:
        """
        
        Gets list of groups a user is in

        Args:
            username (str): The username

        Returns:
            ListUserMembershipResponse: The list of groups
        """
        return await self._auth_client.admin.get_list_group_membership(username=username)

    async def get_check_user_membership(self, username: str, group_id: str) -> CheckMembershipResponse:
        """
        
        Checks user membership within a group

        Args:
            username (str): The username to target
            group_id (str): The group to check

        Returns:
            CheckMembershipResponse: Response
        """
        return await self._auth_client.admin.get_check_user_membership(username=username, group_id=group_id)

    async def post_groups_add_member(self, group_id: str, user: GroupUser) -> AddMemberResponse:
        """
        
        Adds a member to a group

        Args:
            group_id (str): Id of group
            user (GroupUser): The user to add

        Returns:
            AddMemberResponse: The response
        """
        return await self._auth_client.admin.post_groups_add_member(user=user, group_id=group_id)

    async def delete_remove_member(self, group_id: str, username: str) -> RemoveMemberResponse:
        """
        
        Removes member from group

        Args:
            group_id (str): The id of group
            username (str): The user to remove

        Returns:
            RemoveMemberResponse: The response
        """
        return await self._auth_client.admin.delete_remove_member(username=username, group_id=group_id)

    async def post_add_group(self, group: UserGroupMetadata) -> AddGroupResponse:
        """
        
        Adds a group/creates group

        Args:
            group (UserGroupMetadata): The group details

        Returns:
            AddGroupResponse: The response
        """
        return await self._auth_client.admin.post_add_group(group=group)

    async def put_update_group(self, group: UserGroupMetadata) -> UpdateGroupResponse:
        """
        Updates group details

        Args:
            group (UserGroupMetadata): The group metadata

        Returns:
            UpdateGroupResponse: The response
        """
        return await self._auth_client.admin.put_update_group(group=group)

    async def get_export_groups(self) -> GroupsExportResponse:
        """
        Exports all group details in specified format.

        Returns:
            GroupsExportResponse: The data dump
        """
        return await self._auth_client.admin.get_export_groups()

    async def post_import_groups(self, body: GroupsImportRequest) -> GroupsImportResponse:
        """
        
        Imports groups from data dump back in

        Args:
            body (GroupsImportRequest): The import request incl. dump

        Returns:
            GroupsImportResponse: The response
        """
        return await self._auth_client.admin.post_import_groups(body=body)

    async def post_restore_groups_from_table(self, table_name: str, body: GroupsRestoreRequest) -> GroupsImportResponse:
        """
        
        Restores groups by first dumping from valid group table. 
        Needs permissions to table.

        Args:
            table_name (str): The table name
            body (GroupsRestoreRequest): The request

        Returns:
            GroupsImportResponse: The response details
        """
        return await self._auth_client.admin.post_restore_groups_from_table(body=body, table_name=table_name)

    async def get_link_lookup_username(self, username: str) -> AdminLinkUserLookupResponse:
        """
        Gets the linked person for user

        Args:
            username (str): User to lookup

        Returns:
            AdminLinkUserLookupResponse: Response
        """
        return await self._auth_client.admin.get_link_lookup_username(username=username)

    async def get_link_reverse_lookup_username(self, person_id: str) -> UserLinkReverseLookupResponse:
        """
        
        Looks up reverse by person ID

        Args:
            person_id (str): The person to lookup

        Returns:
            UserLinkReverseLookupResponse: The response
        """
        return await self._auth_client.admin.get_link_reverse_lookup_username(person_id=person_id)

    async def post_link_assign(self, body: AdminLinkUserAssignRequest) -> AdminLinkUserAssignResponse:
        """
        
        Assigns a person to a given user

        Args:
            body (AdminLinkUserAssignRequest): The request

        Returns:
            AdminLinkUserAssignResponse: The response
        """
        return await self._auth_client.admin.post_link_assign(body=body)

    async def delete_clear_link(self, username: str) -> AdminLinkUserClearResponse:
        """
        Deletes an existing link

        Args:
            username (str): The user to unlink

        Returns:
            AdminLinkUserClearResponse: The response
        """
        return await self._auth_client.admin.delete_clear_link(username=username)


class Auth(ModuleService):
    # Internal clients
    _auth_client: AuthClient

    # Sub Modules
    admin: AdminAuthSubModule

    def __init__(self, auth: AuthManager, config: Config, auth_client: AuthClient) -> None:
        """

        Sets up the auth API L3 module. 

        Includes admin sub module.

        Parameters
        ----------
        auth : AuthManager
            An abstract interface containing the user's requested auth flow
            method.
        config : Config
            A config object which contains information related to the Provena
            instance. 
        """
        self._auth = auth
        self._config = config

        # Clients related to the datastore scoped as private.
        self._auth_client = auth_client

        # Sub Modules
        self.admin = AdminAuthSubModule(
            auth=self._auth,
            config=self._config,
            auth_client=self._auth_client
        )

    async def get_health_check(self) -> HealthCheckResponse:
        """
        Health check the API

        Returns:
            HealthCheckResponse: Response
        """
        return await self._auth_client.get_health_check()

    async def get_user_request_history(self) -> AccessRequestList:
        """
        
        Gets the users access request history

        Returns:
            AccessRequestList: The list of requests
        """
        return await self._auth_client.get_user_request_history()

    async def post_user_request_change(self, body: AccessReport, send_email: bool) -> StatusResponse:
        """
        
        Requests a change by diffing access models

        Args:
            body (AccessReport): The new access desired
            send_email (bool): Email alert

        Returns:
            StatusResponse: Ok?
        """
        return await self._auth_client.post_user_request_change(body=body, send_email=send_email)

    async def get_user_pending_request_history(self) -> AccessRequestList:
        """
        Gets only pending requests from history

        Returns:
            AccessRequestList: The list
        """
        return await self._auth_client.get_user_pending_request_history()

    async def get_user_generate_access_report(self) -> AccessReportResponse:
        """
        
        Generates an access report detailing system access.

        Returns:
            AccessReportResponse: The response
        """
        return await self._auth_client.get_user_generate_access_report()

    async def get_list_groups(self) -> ListGroupsResponse:
        """
        Lists all groups

        Returns:
            ListGroupsResponse: List of groups
        """
        return await self._auth_client.get_list_groups()

    async def get_describe_group(self, group_id: str) -> DescribeGroupResponse:
        """
        
        Describes a specific gruop

        Args:
            group_id (str): The id of group

        Returns:
            DescribeGroupResponse: Response with details
        """
        return await self._auth_client.get_describe_group(group_id=group_id)

    async def get_list_membership(self) -> ListUserMembershipResponse:
        """
        
        Gets the list of groups user is member of

        Returns:
            ListUserMembershipResponse: List and details
        """
        return await self._auth_client.get_list_membership()

    async def get_list_group_members(self, group_id: str) -> ListMembersResponse:
        """
        
        Lists the members of a given group

        Args:
            group_id (str): The group to lookup

        Returns:
            ListMembersResponse: Members
        """
        return await self._auth_client.get_list_group_members(group_id=group_id)

    async def get_check_membership(self, group_id: str) -> CheckMembershipResponse:
        """
        Checks if user is in specific group

        Args:
            group_id (str): The group to check

        Returns:
            CheckMembershipResponse: In group?
        """
        return await self._auth_client.get_check_membership(group_id=group_id)

    async def get_link_lookup_username(self, username: Optional[str] = None) -> UserLinkUserLookupResponse:
        """
        
        Looks up either current user or specified user

        Args:
            username (Optional[str], optional): The username if not current user. Defaults to None.

        Returns:
            UserLinkUserLookupResponse: The response indicating link
        """
        return await self._auth_client.get_link_lookup_username(username=username)

    async def post_link_assign(self, body: UserLinkUserAssignRequest) -> UserLinkUserAssignResponse:
        """
        
        Assigns link to current user.

        Args:
            body (UserLinkUserAssignRequest): The link to assign

        Returns:
            UserLinkUserAssignResponse: The response indicating success
        """
        return await self._auth_client.post_link_assign(body=body)

    async def post_link_validate(self, body: UserLinkUserAssignRequest) -> UserLinkUserValidateResponse:
        """
        Validates link before making it.

        Args:
            body (UserLinkUserAssignRequest): The link to assign

        Returns:
            UserLinkUserValidateResponse: Valid?
        """
        return await self._auth_client.post_link_validate(body=body)
